# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math
from collections.abc import Generator
from typing import Any, TypeVar, cast, override

from faebryk.libs.sets.numeric_sets import (
    Number,
    NumberLike,
    NumberLikeR,
    Numeric_Interval,
    Numeric_Interval_Disjoint,
)
from faebryk.libs.sets.sets import BoolSet, P_Set, P_UnitSet
from faebryk.libs.units import (
    HasUnit,
    Quantity,
    Unit,
    dimensionless,
    quantity,
    to_si_str,
)
from faebryk.libs.util import (
    cast_assert,
    not_none,
    once,
    operator_type_check,
    round_str,
)

logger = logging.getLogger(__name__)

# Types --------------------------------------------------------------------------------


QuantityT = TypeVar(
    "QuantityT",
    int,
    float,
    Quantity,
    # Unit,
    contravariant=False,
    covariant=False,
)

type QuantityLike = Quantity | NumberLike  # | Unit
QuantityLikeR = (Quantity, *NumberLikeR)


type Numeric = int | float

Quantity_Interval_DisjointT = TypeVar(
    "Quantity_Interval_DisjointT", bound="Quantity_Interval_Disjoint"
)

type QuantitySetLike = Quantity_Set | QuantityLike | tuple[QuantityLike, QuantityLike]

# Helpers ------------------------------------------------------------------------------


def base_units(units: Unit) -> Unit:
    return cast(Unit, Quantity(1, units).to_base_units().units)


# --------------------------------------------------------------------------------------


class Quantity_Set(P_UnitSet[QuantityLike]):
    def __init__(self, units: Unit):
        super().__init__()
        self.units = units
        self.interval_units = base_units(units)

    def base_to_units(self, value: Number) -> Quantity:
        return cast_assert(
            Quantity, quantity(value, self.interval_units).to(self.units)
        )

    def _format_number(self, number: Number, num_decimals: int = 12) -> str:
        if self.units.is_compatible_with(dimensionless):
            if math.isinf(number):
                return "∞" if number > 0 else "-∞"
            if number == 0:
                return "0"
            rel_dif = abs((number - round(number)) / number)
            if rel_dif < 1e-6:
                return round_str(number, 0)
            return round_str(number, num_decimals)

        # ignore num_decimals because si prefixes scale the number
        return to_si_str(self.base_to_units(number), self.units, num_decimals=3)

    @override
    def serialize_pset(self) -> dict:
        return Quantity_Interval_Disjoint.from_value(self).serialize_pset()

    @override
    @classmethod
    def deserialize_pset(cls, data: dict):
        return Quantity_Interval_Disjoint.deserialize(data)


QuantitySetLikeR = (Quantity_Set, *QuantityLikeR)


class Quantity_Interval(Quantity_Set):
    """
    Continuous interval (min < max) with min, max having the same units.
    """

    def __init__(
        self,
        min: QuantityLike | None = None,
        max: QuantityLike | None = None,
        units: Unit | None = None,
    ):
        min_unit = None
        if min is not None:
            min_unit = HasUnit.get_units_or_dimensionless(min)
            if units and not min_unit.is_compatible_with(units):
                raise ValueError("min incompatible with units")
        max_unit = None
        if max is not None:
            max_unit = HasUnit.get_units_or_dimensionless(max)
            if units and not max_unit.is_compatible_with(units):
                raise ValueError("max incompatible with units")

        if (
            min_unit is not None
            and max_unit is not None
            and not min_unit.is_compatible_with(max_unit)
        ):
            raise ValueError("min and max must have the same units")

        super().__init__(not_none(units or min_unit or max_unit))

        if isinstance(min, Quantity):
            num_min = min.to_base_units().magnitude
            if not isinstance(num_min, (float, int, Number)):
                raise ValueError("min must be a float or int quantity")
        else:
            num_min = min

        if isinstance(max, Quantity):
            num_max = max.to_base_units().magnitude
            if not isinstance(num_max, (float, int, Number)):
                raise ValueError("max must be a float or int quantity")
        else:
            num_max = max

        is_int = isinstance(num_min, int) and isinstance(num_max, int)
        if not is_int:
            num_min = float(num_min) if num_min is not None else float("-inf")
            num_max = float(num_max) if num_max is not None else float("inf")
        assert num_min is not None and num_max is not None

        self._interval = Numeric_Interval(num_min, num_max)

    @staticmethod
    def from_center(center: QuantityLike, abs_tol: QuantityLike) -> "Quantity_Interval":
        if isinstance(center, float):
            center = Number(center)
        if isinstance(center, Quantity) and not isinstance(center.magnitude, Number):
            center = Quantity(Number(center.magnitude), center.units)  # type: ignore
            assert isinstance(center, Quantity)
        if isinstance(abs_tol, float):
            abs_tol = Number(abs_tol)
        if isinstance(abs_tol, Quantity) and not isinstance(abs_tol.magnitude, Number):
            abs_tol = Quantity(Number(abs_tol.magnitude), abs_tol.units)  # type: ignore
            assert isinstance(abs_tol, Quantity)
        left = cast_assert(QuantityLikeR, center - abs_tol)
        right = cast_assert(QuantityLikeR, center + abs_tol)
        return Quantity_Interval(left, right)

    @staticmethod
    def from_center_rel(
        center: QuantityLike, rel_tol: float | Number
    ) -> "Quantity_Interval":
        if isinstance(center, float):
            center = Number(center)
        if isinstance(center, Quantity) and not isinstance(center.magnitude, Number):
            center = Quantity(Number(center.magnitude), center.units)  # type: ignore
            assert isinstance(center, Quantity)
        if isinstance(rel_tol, float):
            rel_tol = Number(rel_tol)
        return Quantity_Interval(
            cast_assert(QuantityLikeR, center - center * rel_tol),
            cast_assert(QuantityLikeR, center + center * rel_tol),
        )

    @staticmethod
    def _from_interval(interval: Numeric_Interval, units: Unit) -> "Quantity_Interval":
        return Quantity_Interval(
            min=quantity(interval._min, base_units(units)),
            max=quantity(interval._max, base_units(units)),
            units=units,
        )

    def as_center_tuple(self, relative: bool = False) -> tuple[QuantityT, QuantityT]:
        center = cast_assert(QuantityLikeR, (self.min_elem + self.max_elem)) / 2
        delta = (self.max_elem - self.min_elem) / 2
        if relative:
            delta /= center
        assert isinstance(center, QuantityLikeR)
        assert isinstance(delta, type(center))
        return center, delta  # type: ignore

    @property
    @once
    def min_elem(self) -> Quantity:
        return self.base_to_units(self._interval.min_elem)

    @property
    @once
    def max_elem(self) -> Quantity:
        return self.base_to_units(self._interval.max_elem)

    def is_empty(self) -> bool:
        return self._interval.is_empty()

    def is_unbounded(self) -> bool:
        return self._interval.is_unbounded()

    @override
    def is_finite(self) -> bool:
        return self._interval.is_finite()

    def is_subset_of(self, other: "Quantity_Interval") -> bool:
        return self._interval.is_subset_of(other._interval)

    def op_intersect_interval(
        self, other: "Quantity_Interval"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            return Quantity_Interval_Disjoint(units=self.units)
        _interval = self._interval.op_intersect_interval(other._interval)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_add_interval(self, other: "Quantity_Interval") -> "Quantity_Interval":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._interval.op_add_interval(other._interval)
        return Quantity_Interval._from_interval(_interval, self.units)

    def op_negate(self) -> "Quantity_Interval":
        _interval = self._interval.op_negate()
        return Quantity_Interval._from_interval(_interval, self.units)

    def op_subtract_interval(self, other: "Quantity_Interval") -> "Quantity_Interval":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._interval.op_subtract_interval(other._interval)
        return Quantity_Interval._from_interval(_interval, self.units)

    def op_mul_interval(self, other: "Quantity_Interval") -> "Quantity_Interval":
        _interval = self._interval.op_mul_interval(other._interval)
        return Quantity_Interval._from_interval(
            _interval, cast(Unit, self.units * other.units)
        )

    def op_invert(self) -> "Quantity_Interval_Disjoint":
        _interval = self._interval.op_invert()
        return Quantity_Interval_Disjoint._from_intervals(_interval, 1 / self.units)

    def op_div_interval(
        self, other: "Quantity_Interval"
    ) -> "Quantity_Interval_Disjoint":
        _interval = self._interval.op_div_interval(other._interval)
        return Quantity_Interval_Disjoint._from_intervals(
            _interval, cast(Unit, self.units / other.units)
        )

    # def __copy__(self) -> Self:
    #    r = Quantity_interval.__new__(Quantity_interval)
    #    r.min = self.min
    #    r.max = self.max
    #    r.empty = self.empty
    #    r.units = self.units
    #    return r

    def __contains__(self, item: Any) -> bool:
        if isinstance(item, (float, int, Number)):
            item = quantity(item)
        if isinstance(item, Quantity):
            if not item.units.is_compatible_with(self.units):
                return False
            item = item.to(self.interval_units).magnitude
        if not isinstance(item, float) and not isinstance(item, int):
            return False
        return self._interval.__contains__(item)

    # yucky with floats
    def __eq__(self, value: Any) -> bool:
        if not HasUnit.check(value):
            return False
        if not self.units.is_compatible_with(value.units):
            return False
        if isinstance(value, Quantity_Interval):
            return self._interval == value._interval
        if (
            isinstance(value, Quantity_Interval_Disjoint)
            and len(value._intervals.intervals) == 1
        ):
            return self._interval == value._intervals.intervals[0]
        return False

    # TODO, convert to base unit first
    def __hash__(self) -> int:
        return hash((self._interval, self.interval_units))

    def __repr__(self) -> str:
        return f"Quantity_Interval({self})"

    def __str__(self) -> str:
        min_ = self._format_number(self._interval._min)
        max_ = self._format_number(self._interval._max)
        if min_ == max_:
            return f"[{min_}]"
        center, rel = self._interval.as_center_rel()
        if rel < 0.5 and round(rel, 2) == rel:
            return f"[{self._format_number(center)} ± {rel * 100:.2f}%]"
        return f"[{min_}, {max_}]"

    # operators
    @operator_type_check
    def __add__(self, other: "Quantity_Interval"):
        return self.op_add_interval(other)

    @operator_type_check
    def __sub__(self, other: "Quantity_Interval"):
        return self.op_subtract_interval(other)

    @operator_type_check
    def __neg__(self):
        return self.op_negate()

    @operator_type_check
    def __mul__(self, other: "Quantity_Interval"):
        return self.op_mul_interval(other)

    @operator_type_check
    def __truediv__(self, other: "Quantity_Interval"):
        return self.op_div_interval(other)

    @operator_type_check
    def __and__(self, other: "Quantity_Interval"):
        return self.op_intersect_interval(other)

    @once
    def is_single_element(self) -> bool:
        return self.min_elem == self.max_elem  # type: ignore #TODO

    @override
    def any(self) -> Quantity:
        return self.min_elem


class Quantity_Singleton(Quantity_Interval):
    """
    Single value.
    Represented by a Quantity interval with min == max.
    """

    def __init__(self, value: QuantityLike):
        # FIXME: handle inf Quantity too
        if value == float("inf") or value == float("-inf"):
            raise ValueError("value cannot be infinity for quantity singleton")

        super().__init__(min=value, max=value)

    def get_value(self) -> Quantity:
        return self.min_elem

    def __iter__(self) -> Generator[Quantity]:
        yield self.min_elem

    @classmethod
    def cast(cls, value: Quantity_Interval) -> "Quantity_Singleton":
        if value.min_elem != value.max_elem:
            raise ValueError(f"Interval is not a singleton: {value}")
        return cls(value.min_elem)


class Quantity_Interval_Disjoint(Quantity_Set):
    """
    Quantity interval (min < max) with gaps. \n
    Represented by Set of multiple Quantity interval (without gaps).
    """

    def __init__(
        self,
        *intervals: "Quantity_Interval | Quantity_Interval_Disjoint | tuple[QuantityLike, QuantityLike]",  # noqa: E501
        units: Unit | None = None,
    ):
        proper_intervals = [
            Quantity_Interval(r[0], r[1]) if isinstance(r, tuple) else r
            for r in intervals
        ]
        interval_units = [
            HasUnit.get_units_or_dimensionless(r) for r in proper_intervals
        ]
        if len(interval_units) == 0 and units is None:
            raise ValueError("units must be provided for empty union")

        super().__init__(units or interval_units[0])

        if not all(self.units.is_compatible_with(u) for u in interval_units):
            raise ValueError("all elements must have compatible units")

        def get_backing(r: "Quantity_Interval | Quantity_Interval_Disjoint"):
            if isinstance(r, Quantity_Interval):
                return r._interval
            else:
                return r._intervals

        self._intervals = Numeric_Interval_Disjoint(
            *(get_backing(r) for r in proper_intervals)
        )

    @classmethod
    def _from_intervals(
        cls: type[Quantity_Interval_DisjointT],
        intervals: "Numeric_Interval_Disjoint",
        units: Unit,
    ) -> Quantity_Interval_DisjointT:
        r = cls.__new__(cls)
        r._intervals = intervals
        r.units = units
        r.interval_units = base_units(units)
        return r

    @classmethod
    def unbounded(
        cls: type[Quantity_Interval_DisjointT], units: Unit
    ) -> Quantity_Interval_DisjointT:
        return cls(Quantity_Interval(units=units))

    def is_empty(self) -> bool:
        return self._intervals.is_empty()

    @property
    @once
    def min_elem(self) -> Quantity:
        if self.is_empty():
            raise ValueError("empty interval cannot have min element")
        return self.base_to_units(self._intervals.min_elem)

    @property
    @once
    def max_elem(self) -> Quantity:
        if self.is_empty():
            raise ValueError("empty interval cannot have max element")
        return self.base_to_units(self._intervals.max_elem)

    def closest_elem(self, target: Quantity) -> Quantity:
        if not self.units.is_compatible_with(target.units):
            raise ValueError("incompatible units")
        return self.base_to_units(
            self._intervals.closest_elem(target.to(self.interval_units).magnitude)
        )

    def is_superset_of(self, other: "Quantity_Interval_Disjoint") -> bool:
        if not self.units.is_compatible_with(other.units):
            return False
        return self._intervals.is_superset_of(
            Quantity_Interval_Disjoint.from_value(other)._intervals
        )

    def is_subset_of(self, other: "Quantity_Interval_Disjoint") -> bool:
        if not self.units.is_compatible_with(other.units):
            return False
        return self._intervals.is_subset_of(
            Quantity_Interval_Disjoint.from_value(other)._intervals
        )

    def op_intersect_interval(
        self, other: "Quantity_Interval"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._intervals.op_intersect_interval(other._interval)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_intersect_intervals(
        self, *other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        # TODO make pretty
        def single(left, right):
            if not left.units.is_compatible_with(right.units):
                raise ValueError("incompatible units")
            _interval = left._intervals.op_intersect_intervals(right._intervals)
            return Quantity_Interval_Disjoint._from_intervals(_interval, left.units)

        out = Quantity_Interval_Disjoint(self)

        for o in other:
            out = single(out, o)

        return out

    def op_union_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._intervals.op_union_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_difference_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._intervals.op_difference_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_symmetric_difference_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._intervals.op_symmetric_difference_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_add_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._intervals.op_add_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_negate(self) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_negate()
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_subtract_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(other.units):
            raise ValueError("incompatible units")
        _interval = self._intervals.op_subtract_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_mul_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_mul_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(
            _interval, cast(Unit, self.units * other.units)
        )

    def op_invert(self) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_invert()
        return Quantity_Interval_Disjoint._from_intervals(_interval, 1 / self.units)

    def op_div_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_div_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(
            _interval, cast(Unit, self.units / other.units)
        )

    def op_pow_intervals(
        self, other: "Quantity_Interval_Disjoint"
    ) -> "Quantity_Interval_Disjoint":
        if not other.units.is_compatible_with(dimensionless):
            raise ValueError("exponent must have dimensionless units")
        if other.min_elem != other.max_elem and not self.units.is_compatible_with(
            dimensionless
        ):
            raise ValueError(
                "base must have dimensionless units when exponent is interval"
            )
        units = self.units**other.min_elem.magnitude
        _interval = self._intervals.op_pow_intervals(other._intervals)
        return Quantity_Interval_Disjoint._from_intervals(_interval, units)

    def op_round(self, ndigits: int = 0) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_round(ndigits)
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_abs(self) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_abs()
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_log(self) -> "Quantity_Interval_Disjoint":
        _interval = self._intervals.op_log()
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_sqrt(self) -> "Quantity_Interval_Disjoint":
        return self**0.5

    def op_sin(self) -> "Quantity_Interval_Disjoint":
        if not self.units.is_compatible_with(dimensionless):
            raise ValueError("sin only defined for dimensionless quantities")
        _interval = self._intervals.op_sin()
        return Quantity_Interval_Disjoint._from_intervals(_interval, self.units)

    def op_cos(self) -> "Quantity_Interval_Disjoint":
        return (self + quantity(math.pi / 2, self.units)).op_sin()

    def op_floor(self) -> "Quantity_Interval_Disjoint":
        return (self - quantity(0.5, self.units)).op_round()

    def op_ceil(self) -> "Quantity_Interval_Disjoint":
        return (self + quantity(0.5, self.units)).op_round()

    def op_total_span(self) -> Quantity:
        """Returns the sum of the spans of all intervals in this disjoint set.
        For a single interval, this is equivalent to max - min.
        For multiple intervals, this sums the spans of each disjoint interval."""
        return quantity(
            sum(abs(r.max_elem - r.min_elem) for r in self._intervals), self.units
        )

    def op_deviation_to(
        self, other: QuantitySetLike, relative: bool = False
    ) -> Quantity:
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        sym_diff = self.op_symmetric_difference_intervals(other_qty)
        deviation = sym_diff.op_total_span()
        if relative:
            deviation /= max(abs(self).max_elem, abs(other_qty).max_elem)
        return deviation

    def __contains__(self, item: Any) -> bool:
        if isinstance(item, (float, int, Number)):
            item = quantity(item)
        if isinstance(item, Quantity):
            if not item.units.is_compatible_with(self.units):
                return False
            item = item.to(self.interval_units).magnitude
        if not isinstance(item, float) and not isinstance(item, int):
            return False
        return self._intervals.__contains__(item)

    @once
    def __hash__(self) -> int:
        return hash((self._intervals, self.interval_units))

    @once
    def __repr__(self) -> str:
        return f"Quantity_Interval_Disjoint({self})"

    @once
    def __str__(self) -> str:
        def _format_interval(r: Numeric_Interval) -> str:
            if r._min == r._max:
                return f"[{self._format_number(r._min)}]"
            try:
                center, rel = r.as_center_rel()
                if rel < 0.5 and round(rel, 2) == rel:
                    return f"[{self._format_number(center)} ± {rel * 100:.2f}%]"
            except ZeroDivisionError:
                pass

            return f"[{self._format_number(r._min)}, {self._format_number(r._max)}]"

        out = ", ".join(_format_interval(r) for r in self._intervals.intervals)

        return f"({out})"

    def __iter__(self) -> Generator[Quantity_Interval]:
        for r in self._intervals.intervals:
            yield Quantity_Interval._from_interval(r, self.units)

    def is_unbounded(self) -> bool:
        return self._intervals.is_unbounded()

    @override
    def is_finite(self) -> bool:
        return self._intervals.is_finite()

    # operators
    @staticmethod
    def from_value(other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        if isinstance(other, Quantity_Interval_Disjoint):
            return other
        if isinstance(other, Quantity_Singleton):
            return Quantity_Set_Discrete(other.get_value())
        if isinstance(other, Quantity_Interval):
            return Quantity_Interval_Disjoint(other)
        if isinstance(other, Quantity):
            return Quantity_Set_Discrete(other)
        if isinstance(other, tuple) and len(other) == 2:
            return Quantity_Interval_Disjoint(other)
        if isinstance(other, NumberLike):
            return Quantity_Set_Discrete(quantity(other))
        raise ValueError(f"unsupported type: {type(other)}")

    @staticmethod
    def intersect_all(*obj: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        if not obj:
            return Quantity_Set_Empty()
        intersected = Quantity_Interval_Disjoint.from_value(obj[0])
        for o in obj[1:]:
            intersected = intersected & Quantity_Interval_Disjoint.from_value(o)
        return intersected

    def __eq__(self, value: Any) -> bool:
        if value is None:
            return False
        if not isinstance(value, QuantitySetLikeR):
            return False
        value_q = Quantity_Interval_Disjoint.from_value(value)
        return self._intervals == value_q._intervals

    def __add__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_add_intervals(other_qty)

    def __radd__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return self + other

    def __sub__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_subtract_intervals(other_qty)

    def __rsub__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return -self + other

    def __neg__(self) -> "Quantity_Interval_Disjoint":
        return self.op_negate()

    def __mul__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_mul_intervals(other_qty)

    def __rmul__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return self * other

    def __truediv__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_div_intervals(other_qty)

    def __rtruediv__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return self.op_invert() * Quantity_Interval_Disjoint.from_value(other)

    def __pow__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return self.op_pow_intervals(Quantity_Interval_Disjoint.from_value(other))

    def __and__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_intersect_intervals(other_qty)

    def __rand__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return Quantity_Interval_Disjoint.from_value(other) & self

    def __or__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_union_intervals(other_qty)

    def __ror__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return Quantity_Interval_Disjoint.from_value(other) | self

    def __xor__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        try:
            other_qty = Quantity_Interval_Disjoint.from_value(other)
        except ValueError:
            return NotImplemented
        return self.op_symmetric_difference_intervals(other_qty)

    def __rxor__(self, other: QuantitySetLike) -> "Quantity_Interval_Disjoint":
        return Quantity_Interval_Disjoint.from_value(other) ^ self

    def __ge__(self, other: QuantitySetLike) -> BoolSet:
        other_q = Quantity_Interval_Disjoint.from_value(other)
        if not self.units.is_compatible_with(other_q.units):
            raise ValueError("incompatible units")
        return self._intervals >= other_q._intervals

    def __gt__(self, other: QuantitySetLike) -> BoolSet:
        other_q = Quantity_Interval_Disjoint.from_value(other)
        if not self.units.is_compatible_with(other_q.units):
            raise ValueError("incompatible units")
        return self._intervals > other_q._intervals

    def __le__(self, other: QuantitySetLike) -> BoolSet:
        other_q = Quantity_Interval_Disjoint.from_value(other)
        if not self.units.is_compatible_with(other_q.units):
            raise ValueError("incompatible units")
        return self._intervals <= other_q._intervals

    def __lt__(self, other: QuantitySetLike) -> BoolSet:
        other_q = Quantity_Interval_Disjoint.from_value(other)
        if not self.units.is_compatible_with(other_q.units):
            raise ValueError("incompatible units")
        return self._intervals < other_q._intervals

    def __round__(self, ndigits: int = 0) -> "Quantity_Interval_Disjoint":
        return self.op_round(ndigits)

    def __abs__(self) -> "Quantity_Interval_Disjoint":
        return self.op_abs()

    @once
    def is_single_element(self) -> bool:
        if self.is_empty():
            return False
        return self.min_elem == self.max_elem  # type: ignore #TODO

    @property
    def is_integer(self) -> bool:
        return all(r.is_integer for r in self._intervals.intervals)

    def as_gapless(self) -> "Quantity_Interval":
        if self.is_empty():
            raise ValueError("empty interval cannot be gapless")
        return Quantity_Interval(self.min_elem, self.max_elem, units=self.units)

    @override
    def any(self) -> Quantity:
        return self.min_elem

    def serialize_pset(self) -> dict[str, Any]:
        return {
            "intervals": self._intervals.serialize(),
            "unit": str(self.units),
        }

    @override
    @classmethod
    def deserialize_pset(cls, data: dict):
        from faebryk.libs.units import P

        out = cls(units=getattr(P, data["unit"]))
        out._intervals = P_Set.deserialize(data["intervals"])
        return out

    def to_dimensionless(self) -> "Quantity_Interval_Disjoint":
        return Quantity_Interval_Disjoint._from_intervals(
            self._intervals, dimensionless
        )


class Quantity_Set_Discrete(Quantity_Interval_Disjoint):
    """
    Quantity Set of single values. \n
    Represented by Set of multiple Quantity Interval
    (each containing only a single value, aka a Quantity Singleton).
    """

    def __init__(self, *values: QuantityLike, units: Unit | None = None):
        super().__init__(*(Quantity_Singleton(v) for v in values), units=units)

    def iter_singles(self) -> Generator[Quantity]:
        for r in self._intervals.intervals:
            yield self.base_to_units(r._min)


def Quantity_Set_Empty(
    units: Unit | None = None,
) -> Quantity_Interval_Disjoint:
    """
    Empty Quantity Set.
    Represented by empty Quantity_Interval_Disjoint.
    """

    if units is None:
        units = dimensionless
    return Quantity_Interval_Disjoint(units=units)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.libs.util import KeyErrorNotFound

logger = logging.getLogger(__name__)


class has_pin_association_heuristic_lookup_table(
    F.has_pin_association_heuristic.impl()
):
    def __init__(
        self,
        mapping: dict[F.Electrical, list[str]],
        accept_prefix: bool,
        case_sensitive: bool,
        nc: list[str] | None = None,
    ) -> None:
        super().__init__()
        self.mapping = mapping
        self.accept_prefix = accept_prefix
        self.case_sensitive = case_sensitive
        self.nc = nc or ["NC", "nc"]

    def get_pins(
        self,
        pins: list[tuple[str, str]],
    ) -> dict[str, F.Electrical]:
        """
        Get the pinmapping for a list of pins based on a lookup table.

        :param pins: A list of tuples with the pin number and name.
        :return: A dictionary with the pin name as key and the module interface as value
        """

        pinmap = {}
        for mif, alt_names in self.mapping.items():
            matches = []
            for number, name in pins:
                if name in self.nc:
                    continue
                for alt_name in alt_names:
                    if not self.case_sensitive:
                        alt_name = alt_name.lower()
                        name = name.lower()
                    if self.accept_prefix and name.endswith(alt_name):
                        matches.append((number, name))
                        break
                    elif name == alt_name:
                        matches.append((number, name))
                        break
            if not matches:
                try:
                    _, is_optional = mif.get_parent_with_trait(F.is_optional)
                    if not is_optional.is_needed():
                        is_optional._handle_result(False)
                        continue
                except KeyErrorNotFound:
                    pass
                raise F.has_pin_association_heuristic.PinMatchException(
                    f"Could not find a match for pin {mif} with names '{alt_names}'"
                    f" in the pins: {pins}"
                )
            for number, name in matches:
                pinmap[number] = mif
                logger.debug(
                    f"Matched pin {number} with name {name} to {mif} with "
                    f"alias {alt_names}"
                )

        unmatched = [p for p in pins if p[0] not in pinmap and p[1] not in self.nc]
        if unmatched:
            logger.warning(
                f"Unmatched pins: {unmatched}, all pins: {pins},"
                f" mapping: {self.mapping}"
            )
        return pinmap



"""
This file defines what attributes are available in `ato`.
"""

import logging
import re
from typing import Type

import faebryk.core.parameter as fab_param
import faebryk.library._F as F
import faebryk.libs.library.L as L
from atopile import address
from atopile.errors import UserBadParameterError, UserNotImplementedError
from faebryk.core.trait import TraitImpl, TraitNotFound
from faebryk.libs.exceptions import downgrade
from faebryk.libs.picker.picker import DescriptiveProperties

log = logging.getLogger(__name__)


shim_map: dict[address.AddrStr, tuple[Type[L.Node], str]] = {}


def _register_shim(addr: str | address.AddrStr, preferred: str):
    def _wrapper[T: Type[L.Node]](cls: T) -> T:
        shim_map[address.AddrStr(addr)] = cls, preferred
        return cls

    return _wrapper


class _has_local_kicad_footprint_named_defined(F.has_footprint_impl):
    """
    This trait defers footprint creation until it's needed,
    which means we can construct the underlying pin map
    """

    def __init__(self, lib_reference: str, pinmap: dict[str, F.Electrical]):
        super().__init__()
        if ":" not in lib_reference:
            # TODO: default to a lib reference starting with "lib:"
            # for backwards compatibility with old footprints
            lib_reference = f"lib:{lib_reference}"
        self.lib_reference = lib_reference
        self.pinmap = pinmap

    def get_footprint(self) -> F.Footprint:
        if fp := self.try_get_footprint():
            return fp
        else:
            fp = F.KicadFootprint(
                pin_names=list(self.pinmap.keys()),
            )
            fp.add(F.KicadFootprint.has_kicad_identifier(self.lib_reference))
            fp.get_trait(F.can_attach_via_pinmap).attach(self.pinmap)  # type: ignore
            self.set_footprint(fp)
            return fp

    def handle_duplicate(
        self, old: "_has_local_kicad_footprint_named_defined", _: fab_param.Node
    ) -> bool:
        if old.try_get_footprint():
            raise RuntimeError("Too late to set footprint")

        # Update the existing trait...
        old.lib_reference = self.lib_reference
        # ... and we don't need to attach the new
        assert old.pinmap is self.pinmap, "Pinmap reference mismatch"
        return False


class _has_ato_cmp_attrs(L.Module.TraitT.decless()):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.pinmap: dict[str, F.Electrical | None] = {}

    def on_obj_set(self):
        self.module = self.get_obj(L.Module)
        self.module.add(F.can_attach_to_footprint_via_pinmap(self.pinmap))
        self.module.add(F.has_designator_prefix(F.has_designator_prefix.Prefix.U))

    def add_pin(self, pinname: str, pyname: str) -> F.Electrical:
        mif = self.module.add(F.Electrical(), name=pyname)

        self.pinmap[pinname] = mif
        return mif

    def handle_duplicate(self, old: TraitImpl, node: fab_param.Node) -> bool:
        # Don't replace the existing ato trait on addition
        return False


# FIXME: this would ideally be some kinda of mixin,
# however, we can't have multiple bases for Nodes
class GlobalAttributes(L.Module):
    """
    These attributes are available to all modules and interfaces in a design.
    """

    @property
    def lcsc_id(self):
        """
        Assign the LCSC ID of the module.

        If set, this will tell the picker to select that part from LCSC for this block.
        """
        raise AttributeError("write-only")

    @lcsc_id.setter
    def lcsc_id(self, value: str):
        # handles duplicates gracefully
        self.add(F.has_descriptive_properties_defined({"LCSC": value}))

    @property
    def manufacturer(self) -> str:
        """
        This module's manufacturer name, as a string.

        Only exact matches on the manufacturer's name will be found by the picker.
        It's recommended to fill this information based on what `ato create component`
        provides.
        """
        raise AttributeError("write-only")

    @manufacturer.setter
    def manufacturer(self, value: str):
        # handles duplicates gracefully
        self.add(
            F.has_descriptive_properties_defined(
                {DescriptiveProperties.manufacturer: value}
            )
        )

    @property
    def mpn(self) -> str:
        """
        This module's manufacturer part number, as a string.

        For the picker to select the correct part from the manufacturer,
        this must be set.
        """
        try:
            return self.get_trait(F.has_descriptive_properties).get_properties()[
                DescriptiveProperties.partno
            ]
        except (TraitNotFound, KeyError):
            raise AttributeError(name="mpn", obj=self)

    @mpn.setter
    def mpn(self, value: str):
        from atopile.front_end import DeprecatedException

        if value.lower() == "dnp":
            raise DeprecatedException(
                f"`mpn = {value}` is deprecated. Use `exclude_from_bom = True` instead."
            )

        # handles duplicates gracefully
        self.add(
            F.has_descriptive_properties_defined({DescriptiveProperties.partno: value})
        )

        # TODO: @v0.4: remove this - mpn != lcsc id
        if re.match(r"C[0-9]+", value):
            self.add(F.has_descriptive_properties_defined({"LCSC": value}))

            raise DeprecatedException(
                "`mpn` is deprecated for assignment of LCSC IDs. Use `lcsc_id` instead."
            )

    @property
    def datasheet_url(self) -> str:
        """
        The URL of the datasheet for this component.
        """
        raise AttributeError("write-only")

    @datasheet_url.setter
    def datasheet_url(self, value: str):
        self.add(F.has_datasheet_defined(value))

    @property
    def designator_prefix(self):
        """
        The prefix used for automatically-generated designators on this module.
        """
        raise AttributeError("write-only")

    @designator_prefix.setter
    def designator_prefix(self, value: str):
        self.add(F.has_designator_prefix(value))

    @property
    def package(self) -> str:
        """
        The package of the module.

        This drives which components can be selected, and what footprint is used.

        Must exactly match a known package name.
        """
        raise AttributeError("write-only")

    @package.setter
    def package(self, value: str):
        try:
            self.add(F.has_package(value))
        except ValueError:
            valid_packages = ", ".join(
                f"`{k}`" for k in F.has_package.Package.__members__.keys()
            )
            raise UserBadParameterError(
                f"Invalid package: `{value}`. Valid packages are: {valid_packages}"
            )

    @property
    def footprint(self) -> str:
        """
        Explicitly set the footprint to be used for this module.

        Setting this will cause this component to be selected and placed on the PCB.

        The footprint should be a string, naming the KiCAD ID of the footprint.
        """
        raise AttributeError("write-only")

    @footprint.setter
    def footprint(self, value: str):
        self.add(
            _has_local_kicad_footprint_named_defined(
                value, self.get_trait(_has_ato_cmp_attrs).pinmap
            )
        )

    # TODO: do not place
    @property
    def exclude_from_bom(self):
        raise AttributeError("write-only")

    # See: https://github.com/atopile/atopile/pull/424/files#diff-63194bff2019ade91098b68b1c47e10ce94fb03258923f8c77f66fc5707a0c96
    @exclude_from_bom.setter
    def exclude_from_bom(self, value: bool):
        raise UserNotImplementedError(
            "`exclude_from_bom` is not yet implemented. "
            "This should not currently affect your design, "
            "however may throw spurious warnings.\n"
            "See: https://github.com/atopile/atopile/issues/755"
        )

    @property
    def override_net_name(self):
        """
        When set on an interface, this will override the net name of the interface.

        This is useful for renaming nets which are automatically generated.
        """
        raise AttributeError("write-only")

    @override_net_name.setter
    def override_net_name(self, name: str):
        self.add(F.has_net_name(name, level=F.has_net_name.Level.EXPECTED))

    @property
    def suggest_net_name(self):
        """
        Suggested net name which will have a higher priority than generated net names.
        """
        raise AttributeError("write-only")

    @suggest_net_name.setter
    def suggest_net_name(self, name: str):
        """
        Suggested net name which will have a higher priority than generated net names.
        """
        self.add(F.has_net_name(name, level=F.has_net_name.Level.SUGGESTED))

    @property
    def required(self):
        """
        Only for ModuleInterfaces.
        If set to `True`, require that interface is connected to something outside
        of the module it's defined in.
        """
        raise AttributeError("write-only")

    @required.setter
    def required(self, value: bool):
        if not value:
            self.del_trait(F.requires_external_usage)
            return
        self.add(F.requires_external_usage())


def _handle_package_shim(module: L.Module, value: str, starts_with: str):
    if (prefixed_value := f"{starts_with}{value}") in F.has_package.Package.__members__:
        value = prefixed_value

    try:
        pkg = F.has_package.Package(value)
    except ValueError:
        valid_packages = ", ".join(
            [
                f"`{k}`"
                for k in F.has_package.Package.__members__.keys()
                if k.startswith(starts_with)
            ]
        )
        raise UserBadParameterError(
            f"Invalid package for {module.__class__.__name__}: `{value}`. "
            f"Valid packages are: {valid_packages}"
        )
    else:
        module.add(F.has_package(pkg))


@_register_shim("generics/resistors.ato:Resistor", "import Resistor")
class Resistor(F.Resistor):
    """
    This resistor is replaces `generics/resistors.ato:Resistor`
    every times it's referenced.
    """

    @property
    def value(self):
        """Represents the resistance of the resistor."""
        return self.resistance

    @value.setter
    def value(self, value: L.Range):
        self.resistance.constrain_subset(value)

    @property
    def footprint(self):
        """See `GlobalAttributes.footprint`"""
        raise AttributeError("write-only")

    @footprint.setter
    def footprint(self, value: str):
        from atopile.front_end import DeprecatedException

        if value.startswith("R"):
            try:
                self.package = value[1:]
            except UserBadParameterError:
                pass
            else:
                with downgrade(DeprecatedException):
                    raise DeprecatedException(
                        "`footprint` is deprecated for assignment of package. "
                        f"Use: `package = '{value[1:]}'`"
                    )
                # Return here, to avoid additionally setting the footprint
                return

        GlobalAttributes.footprint.fset(self, value)

    @property
    def package(self):
        """See `GlobalAttributes.package`"""
        raise AttributeError("write-only")

    @package.setter
    def package(self, value: str):
        _handle_package_shim(self, value, "R")

    @property
    def _1(self) -> F.Electrical:
        return self.unnamed[0]

    @property
    def _2(self) -> F.Electrical:
        return self.unnamed[1]

    @L.rt_field
    def has_ato_cmp_attrs_(self) -> _has_ato_cmp_attrs:
        """Ignore this field."""
        trait = _has_ato_cmp_attrs()
        trait.pinmap["1"] = self.p1
        trait.pinmap["2"] = self.p2
        return trait


class CommonCapacitor(F.Capacitor):
    """
    These attributes are common to both electrolytic and non-electrolytic capacitors.
    """

    @property
    def value(self):
        """Represents the capacitance of the capacitor."""
        return self.capacitance

    @value.setter
    def value(self, value: L.Range):
        self.capacitance.constrain_subset(value)

    @property
    def package(self):
        """See `GlobalAttributes.package`"""
        raise AttributeError("write-only")

    @package.setter
    def package(self, value: str):
        _handle_package_shim(self, value, "C")

    @property
    def footprint(self):
        """See `GlobalAttributes.footprint`"""
        raise AttributeError("write-only")

    @footprint.setter
    def footprint(self, value: str):
        from atopile.front_end import DeprecatedException

        if value.startswith("C"):
            try:
                self.package = value[1:]
            except UserBadParameterError:
                pass
            else:
                with downgrade(DeprecatedException):
                    raise DeprecatedException(
                        "`footprint` is deprecated for assignment of package. "
                        f"Use: `package = '{value[1:]}'`"
                    )
                # Return here, to avoid additionally setting the footprint
                return

        GlobalAttributes.footprint.fset(self, value)

    @property
    def _1(self) -> F.Electrical:
        return self.unnamed[0]

    @property
    def _2(self) -> F.Electrical:
        return self.unnamed[1]


@_register_shim("generics/capacitors.ato:Capacitor", "import Capacitor")
class Capacitor(CommonCapacitor):
    """
    This capacitor is replaces `generics/capacitors.ato:Capacitor`
    every times it's referenced.
    """

    @L.rt_field
    def has_ato_cmp_attrs_(self) -> _has_ato_cmp_attrs:
        """Ignore this field."""
        trait = _has_ato_cmp_attrs()
        trait.pinmap["1"] = self.p1
        trait.pinmap["2"] = self.p2
        return trait


@_register_shim(
    "generics/capacitors.ato:CapacitorElectrolytic", "import CapacitorElectrolytic"
)
class CapacitorElectrolytic(CommonCapacitor):
    """Temporary shim to translate capacitors."""

    anode: F.Electrical
    cathode: F.Electrical

    pickable = None

    # Overrides the default implementation in F.Capacitor
    @property
    def power(self) -> F.ElectricPower:
        if self.has_trait(self._has_power):
            power = self.get_trait(self._has_power).power
        else:
            power = F.ElectricPower()
            power.hv.connect(self.anode)
            power.lv.connect(self.cathode)
            self.add(self._has_power(power))

        return power


@_register_shim("generics/inductors.ato:Inductor", "import Inductor")
class Inductor(F.Inductor):
    """
    This inductor is replaces `generics/inductors.ato:Inductor`
    every times it's referenced.
    """

    @property
    def _1(self) -> F.Electrical:
        return self.unnamed[0]

    @property
    def _2(self) -> F.Electrical:
        return self.unnamed[1]

    @L.rt_field
    def has_ato_cmp_attrs_(self) -> _has_ato_cmp_attrs:
        """Ignore this field."""
        trait = _has_ato_cmp_attrs()
        trait.pinmap["1"] = self.p1
        trait.pinmap["2"] = self.p2
        return trait

    @property
    def package(self):
        """See `GlobalAttributes.package`"""
        raise AttributeError("write-only")

    @package.setter
    def package(self, value: str):
        _handle_package_shim(self, value, "L")


@_register_shim("generics/leds.ato:LED", "import LED")
class LED(F.LED):
    """Temporary shim to translate LEDs."""

    @property
    def v_f(self):
        return self.forward_voltage

    @property
    def i_max(self):
        return self.max_current


@_register_shim("generics/interfaces.ato:Power", "import ElectricPower")
class Power(F.ElectricPower):
    """Temporary shim to translate `value` to `power`."""

    @property
    def current(self):
        """
        Maximum current the power interface can provide.

        Negative is current draw.
        """
        return self.max_current


@_register_shim("generics/interfaces.ato:I2C", "import I2C")
class I2C(F.I2C):
    """Temporary shim to translate I2C interfaces."""

    @property
    def gnd(self):
        return self.single_electric_reference.get_reference().gnd



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

logger = logging.getLogger(__name__)


from faebryk.core.cpp import (  # noqa: E402, F401
    Link,
    LinkDirect,
    LinkDirectConditional,
    LinkDirectConditionalFilterResult,
    LinkDirectDerived,
    LinkFilteredException,
    LinkNamedParent,
    LinkParent,
    LinkPointer,
    LinkSibling,
)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""
Runner to use when running under a different interpreter.
"""

import os
import pathlib
import sys
import traceback


# **********************************************************
# Update sys.path before importing any bundled libraries.
# **********************************************************
def update_sys_path(path_to_add: str, strategy: str) -> None:
    """Add given path to `sys.path`."""
    if path_to_add not in sys.path and os.path.isdir(path_to_add):
        if strategy == "useBundled":
            sys.path.insert(0, path_to_add)
        elif strategy == "fromEnvironment":
            sys.path.append(path_to_add)


# Ensure that we can import LSP libraries, and other bundled libraries.
update_sys_path(
    os.fspath(pathlib.Path(__file__).parent.parent / "libs"),
    os.getenv("LS_IMPORT_STRATEGY", "useBundled"),
)


# pylint: disable=wrong-import-position,import-error
import lsp_jsonrpc as jsonrpc  # noqa: E402
import lsp_utils as utils  # noqa: E402

RPC = jsonrpc.create_json_rpc(sys.stdin.buffer, sys.stdout.buffer)

EXIT_NOW = False
while not EXIT_NOW:
    msg = RPC.receive_data()

    method = msg["method"]
    if method == "exit":
        EXIT_NOW = True
        continue

    if method == "run":
        is_exception = False
        # This is needed to preserve sys.path, pylint modifies
        # sys.path and that might not work for this scenario
        # next time around.
        with utils.substitute_attr(sys, "path", sys.path[:]):
            try:
                # TODO: `utils.run_module` is equivalent to running `python -m atopile`.
                # If your tool supports a programmatic API then replace the function below # noqa: E501  # pre-existing
                # with code for your tool. You can also use `utils.run_api` helper, which # noqa: E501  # pre-existing
                # handles changing working directories, managing io streams, etc.
                # Also update `_run_tool_on_document` and `_run_tool` functions in `lsp_server.py`. # noqa: E501  # pre-existing
                result = utils.run_module(
                    module=msg["module"],
                    argv=msg["argv"],
                    use_stdin=msg["useStdin"],
                    cwd=msg["cwd"],
                    source=msg["source"] if "source" in msg else None,
                )
            except Exception:  # pylint: disable=broad-except
                result = utils.RunResult("", traceback.format_exc(chain=True))
                is_exception = True

        response = {"id": msg["id"]}
        if result.stderr:
            response["error"] = result.stderr
            response["exception"] = is_exception
        elif result.stdout:
            response["result"] = result.stdout

        RPC.send_data(response)



import os
import shutil
import subprocess
import sys
import textwrap
from functools import partial
from pathlib import Path
from typing import Callable

import pytest

from faebryk.libs.util import run_live

HERE = Path(__file__).parent
DEFAULT_CONFIG = HERE / "default_ato.yaml"


def exec_build(args: list[str], cwd: Path) -> tuple[str, str, subprocess.Popen]:
    return run_live(
        [sys.executable, "-m", "atopile", "build", *args],
        env={**os.environ, "NONINTERACTIVE": "1"},
        stdout=print,
        stderr=print,
        check=False,
        cwd=cwd,
    )


EXEC_T = Callable[[str, list[str]], tuple[str, str, subprocess.Popen]]


def dump_and_run(src: str, args: list[str], working_dir: Path):
    # Copy the default_ato.yaml to the tmpdir
    shutil.copy2(DEFAULT_CONFIG, working_dir / "ato.yaml")

    with open(working_dir / "app.ato", "w", encoding="utf-8") as f:
        f.write(textwrap.dedent(src))

    return exec_build(args, working_dir)


@pytest.fixture
def build_app(tmpdir: Path):
    yield partial(dump_and_run, working_dir=tmpdir)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P


class TI_CD4011BE(F.CD4011):
    footprint = L.f_field(F.has_footprint_defined)(
        F.DIP(pin_cnt=14, spacing=7.62 * P.mm, long_pads=False)
    )

    def __preinit__(self):
        self.footprint.get_footprint().get_trait(F.can_attach_via_pinmap).attach(
            {
                "7": self.power.lv,
                "14": self.power.hv,
                "3": self.gates[0].outputs[0].line,
                "4": self.gates[1].outputs[0].line,
                "11": self.gates[2].outputs[0].line,
                "10": self.gates[3].outputs[0].line,
                "1": self.gates[0].inputs[0].line,
                "2": self.gates[0].inputs[1].line,
                "5": self.gates[1].inputs[0].line,
                "6": self.gates[1].inputs[1].line,
                "12": self.gates[2].inputs[0].line,
                "13": self.gates[2].inputs[1].line,
                "9": self.gates[3].inputs[0].line,
                "8": self.gates[3].inputs[1].line,
            }
        )

        self.add(
            F.has_descriptive_properties_defined(
                {
                    DescriptiveProperties.manufacturer: "Texas Instruments",
                    DescriptiveProperties.partno: "CD4011BE",
                },
            )
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P
from faebryk.libs.util import not_none  # noqa: F401

logger = logging.getLogger(__name__)


class CH344(Module):
    """
    Quad UART to USB bridge
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    usb: F.USB2_0_IF.Data
    uart = L.list_field(4, F.UART)
    act: F.ElectricLogic
    indicator_tx: F.ElectricLogic
    indicator_rx: F.ElectricLogic
    osc = L.list_field(2, F.Electrical)
    reset: F.ElectricLogic
    test: F.ElectricLogic
    cfg: F.ElectricLogic
    gpio = L.list_field(16, F.ElectricLogic)
    power: F.ElectricPower

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            not_none(F.ElectricLogic.connect_all_module_references(self))
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wch-ic.com/downloads/CH344DS1_PDF.html"
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.gpio[0].connect(self.uart[0].cts)
        self.gpio[1].connect(self.uart[0].rts)
        self.gpio[2].connect(self.uart[1].cts)
        self.gpio[3].connect(self.uart[1].rts)
        self.gpio[4].connect(self.uart[2].cts)
        self.gpio[5].connect(self.uart[2].rts)
        self.gpio[6].connect(self.uart[3].cts)
        self.gpio[7].connect(self.uart[3].rts)
        self.gpio[8].connect(self.uart[0].dtr)
        self.gpio[9].connect(self.uart[1].dtr)
        self.gpio[10].connect(self.uart[2].dtr)
        self.gpio[11].connect(self.uart[3].dtr)
        self.gpio[12].connect(self.uart[0].dcd)
        self.gpio[13].connect(self.uart[0].ri)
        self.gpio[14].connect(self.uart[0].dsr)
        self.gpio[15].connect(self.uart[1].dcd)

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power.voltage.constrain_subset(L.Range.from_center_rel(3.3 * P.V, 0.1))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import pprint
from dataclasses import dataclass
from textwrap import indent
from typing import Callable, Iterable

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.parameter import (
    And,
    Is,
    Or,
    Parameter,
    ParameterOperatable,
    Predicate,
)
from faebryk.core.solver.solver import Solver
from faebryk.core.solver.utils import Contradiction
from faebryk.libs.picker.picker import DescriptiveProperties, Part, PickError
from faebryk.libs.util import not_none

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PickerOption:
    part: Part
    params: dict[str, ParameterOperatable.SetLiteral] | None = None
    """
    Parameters that need to be matched for this option to be valid.

    Assumes specified params are narrowest possible value for this part
    """
    filter: Callable[[Module], bool] | None = None
    pinmap: dict[str, F.Electrical] | None = None
    info: dict[str | DescriptiveProperties, str] | None = None

    def __hash__(self):
        return hash(self.part)


class PickErrorParams(PickError):
    def __init__(self, module: Module, options: list[PickerOption], solver: Solver):
        self.options = options

        MAX = 5

        options_str = "\n".join(
            f"{pprint.pformat(o.params, indent=4)}" for o in self.options[:MAX]
        )
        if len(self.options) > MAX:
            options_str += f"\n... and {len(self.options) - MAX} more"

        message = (
            f"Could not find part for {module}"
            f"\nwith params:\n{indent(module.pretty_params(solver), ' '*4)}"
            f"\nin options:\n {indent(options_str, ' '*4)}"
        )
        super().__init__(message, module)


def pick_module_by_params(
    module: Module, solver: Solver, options: Iterable[PickerOption]
):
    if module.has_trait(F.has_part_picked):
        logger.debug(f"Ignoring already picked module: {module}")
        return

    params = {
        not_none(p.get_parent())[1]: p
        for p in module.get_children(direct_only=True, types=Parameter)
    }

    filtered_options = [o for o in options if not o.filter or o.filter(module)]
    option: PickerOption | None = None
    contradictions = []

    for o in filtered_options:
        predicate_list: list[Predicate] = []

        for k, v in (o.params or {}).items():
            if not k.startswith("_"):
                param = params[k]
                predicate_list.append(Is(param, v))

        # No predicates, thus always valid option
        if len(predicate_list) == 0:
            predicate = Or(True)
            continue

        predicate = And(*predicate_list)
        try:
            solver.try_fulfill(predicate, lock=True)
        except Contradiction as c:
            contradictions.append(c)
        else:
            option = o
            break

    if option is None:
        # TODO pass the contradictions
        raise PickErrorParams(module, list(options), solver)

    if option.pinmap:
        module.add(F.can_attach_to_footprint_via_pinmap(option.pinmap))

    option.part.supplier.attach(module, option)
    module.add(F.has_part_picked(option.part))

    logger.debug(f"Attached {option.part.partno} to {module}")
    return option



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from typing import TYPE_CHECKING, TypeGuard, cast

from faebryk.core.node import Node, NodeException
from faebryk.libs.util import cast_assert

logger = logging.getLogger(__name__)


class TraitNotFound(NodeException):
    def __init__(self, node: Node, trait: type["Trait"], *args: object) -> None:
        super().__init__(
            node,
            *args,
            f"Trait `{trait.__qualname__}` not found in "
            f"`{type(node).__qualname__}[{node}]`",
        )
        self.trait = trait


class TraitImplementationConfusedWithTrait(NodeException):
    def __init__(self, node: Node, trait: type["Trait"], *args: object) -> None:
        super().__init__(
            node,
            *args,
            "Implementation or trait was used where the other was expected.",
        )
        self.trait = trait


class TraitAlreadyExists(NodeException):
    def __init__(self, node: Node, trait: "TraitImpl", *args: object) -> None:
        trait_type = trait.__trait__
        super().__init__(
            node,
            *args,
            f"Trait `{trait_type.__qualname__}` already exists in "
            f"`{node}`: `{node.get_trait(trait_type)}`"
            f", trying to add `{trait.__qualname__}`",
        )
        self.trait = trait


class TraitUnbound(NodeException):
    def __init__(self, node: Node, *args: object) -> None:
        super().__init__(node, *args, f"Trait `{node}` is not bound to a node")


class Trait(Node):
    __decless_trait__: bool = False

    if TYPE_CHECKING:
        TraitT: type["TraitT"]

    # TODO once
    @classmethod
    def impl[T: "Trait"](cls: type[T]):
        # Implements TraitImpl
        class _Impl(cls):
            __trait__: type[Trait] = cls

            @property
            def obj(self) -> Node:
                p = self.get_parent()
                if not p:
                    raise TraitUnbound(self)
                return cast_assert(Node, p[0])

            def get_obj[TN: Node](self, type: type[TN]) -> TN:
                return cast_assert(type, self.obj)

            def cmp(self, new_t: "TraitImpl") -> tuple[bool, "TraitImpl"]:
                assert TraitImpl.is_traitimpl(new_t)

                # If new same or more specific
                if new_t.implements(self.__trait__):
                    return True, new_t

                # hack type (ghetto protocol)
                traitimpl = cast(TraitImpl, self)

                # If we are more specific
                if self.implements(new_t.__trait__):
                    return True, traitimpl

                return False, traitimpl

            def implements(self, trait: type[Trait]):
                return trait.is_traitimpl(self)

            # Overwriteable ------------------------------------------------------------

            def _handle_added_to_parent(self):
                self.on_obj_set()

            def on_obj_set(self): ...

            def handle_duplicate(self, old: "TraitImpl", node: Node) -> bool:
                assert old is not self
                _, candidate = old.cmp(cast(TraitImpl, self))
                if candidate is not self:
                    # raise TraitAlreadyExists(node, self)
                    return False

                node.del_trait(old.__trait__)
                return True

            # override this to implement a dynamic trait
            def is_implemented(self):
                return True

        # this should be outside the class def to prevent
        # __init_subclass__ from overwriting it
        _Impl.__trait__ = cls
        _Impl.__name__ = f"{cls.__name__}Impl"

        return _Impl

    def __new__(cls, *args, **kwargs):
        if not TraitImpl.is_traitimpl_type(cls):
            raise TypeError(f"Don't instantiate Trait [{cls}] use Trait.impl instead")

        return super().__new__(cls)  # type: ignore

    @classmethod
    def decless(cls):
        class _Trait(cls): ...

        _Trait.__decless_trait__ = True

        return _Trait.impl()

    @classmethod
    def is_traitimpl(cls, obj: "Trait") -> TypeGuard["TraitImpl"]:
        assert issubclass(cls, Trait)
        if not TraitImpl.is_traitimpl(obj):
            return False
        return issubclass(obj.__trait__, cls)

    # TODO check subclasses implementing abstractmethods (use subclass_init)


# Hack, using this as protocol
# Can't use actual protocol because CNode doesn't allow multiple inheritance
class TraitImpl(Node):
    """
    Warning: Do not instancecheck against this type!
    """

    __trait__: type[Trait]

    @property
    def obj(self) -> Node: ...
    def get_obj[T: Node](self, type: type[T]) -> T: ...
    def cmp(self, other: "TraitImpl") -> tuple[bool, "TraitImpl"]: ...
    def implements(self, trait: type[Trait]): ...

    # Overwriteable --------------------------------------------------------------------
    def _handle_added_to_parent(self): ...
    def on_obj_set(self): ...
    def handle_duplicate(self, old: "TraitImpl", node: Node) -> bool:
        """
        Returns True if the duplicate was handled, False if the trait should be skipped
        """
        ...

    def is_implemented(self): ...

    @staticmethod
    def is_traitimpl(obj) -> TypeGuard["TraitImpl"]:
        if not isinstance(obj, Trait):
            return False
        return getattr(obj, "__trait__", None) is not None

    @staticmethod
    def is_traitimpl_type(obj) -> TypeGuard[type["TraitImpl"]]:
        if not issubclass(obj, Trait):
            return False
        return getattr(obj, "__trait__", None) is not None


class TraitT(Trait): ...


Trait.TraitT = TraitT



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.units import P


class USB2_0_IF(ModuleInterface):
    class Data(F.DifferentialPair):
        # FIXME: this should be in diffpair right?
        @L.rt_field
        def single_electric_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

        def __preinit__(self):
            self.single_electric_reference.get_reference().voltage.constrain_subset(
                L.Range(0 * P.V, 3.6 * P.V)
            )

    d: Data
    buspower: F.ElectricPower



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from enum import StrEnum
from textwrap import indent
from typing import TYPE_CHECKING, Iterable

import faebryk.library._F as F
from faebryk.core.cpp import Graph
from faebryk.core.graph import GraphFunctions
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.parameter import (
    ConstrainableExpression,
    Is,
    Parameter,
    ParameterOperatable,
)
from faebryk.core.solver.solver import LOG_PICK_SOLVE, Solver
from faebryk.core.solver.utils import Contradiction, ContradictionByLiteral, get_graphs
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.test.times import Times
from faebryk.libs.util import (
    Advancable,
    ConfigFlag,
    EquivalenceClasses,
    KeyErrorAmbiguous,
    KeyErrorNotFound,
    Tree,
    indented_container,
    partition,
    try_or,
    unique,
)

if TYPE_CHECKING:
    from faebryk.libs.picker.api.models import Component
    from faebryk.libs.picker.localpick import PickerOption


NO_PROGRESS_BAR = ConfigFlag("NO_PROGRESS_BAR", default=False)

logger = logging.getLogger(__name__)


class DescriptiveProperties(StrEnum):
    manufacturer = "Manufacturer"
    partno = "Partnumber"
    datasheet = "Datasheet"


class Supplier(ABC):
    @abstractmethod
    def attach(self, module: Module, part: "PickerOption"): ...


@dataclass(frozen=True)
class Part:
    partno: str
    supplier: Supplier


class PickError(Exception):
    def __init__(self, message: str, *modules: Module):
        self.message = message
        self.modules = modules

    def __repr__(self):
        return f"{type(self).__name__}({self.modules}, {self.message})"

    def __str__(self):
        return self.message


class PickErrorNotImplemented(PickError):
    def __init__(self, module: Module):
        message = f"Could not pick part for {module}: Not implemented"
        super().__init__(message, module)


class PickVerificationError(PickError):
    def __init__(self, message: str, *modules: Module):
        message = f"Post-pick verification failed for picked parts:\n{message}"
        super().__init__(message, *modules)


class PickErrorChildren(PickError):
    def __init__(self, module: Module, children: dict[Module, PickError]):
        self.children = children

        message = f"Could not pick parts for children of {module}:\n" + "\n".join(
            f"{m}: caused by {v.message}" for m, v in self.get_all_children().items()
        )
        super().__init__(message, module)

    def get_all_children(self):
        return {
            k: v
            for k, v in self.children.items()
            if not isinstance(v, PickErrorChildren)
        } | {
            module: v2
            for v in self.children.values()
            if isinstance(v, PickErrorChildren)
            for module, v2 in v.get_all_children().items()
        }


class NotCompatibleException(Exception):
    def __init__(
        self,
        module: Module,
        component: "Component",
        param: Parameter | None = None,
        c_range: P_Set | None = None,
    ):
        self.module = module
        self.component = component
        self.param = param
        self.c_range = c_range

        if param is None or c_range is None:
            msg = f"{component.lcsc_display} is not compatible with `{module}`"
        else:
            msg = (
                f"`{param}` ({param.try_get_literal_subset()}) is not "
                f"compatible with {component.lcsc_display} ({c_range})"
            )

        super().__init__(msg)


class does_not_require_picker_check(Parameter.TraitT.decless()):
    pass


def get_pick_tree(module: Module | ModuleInterface) -> Tree[Module]:
    if isinstance(module, Module):
        module = module.get_most_special()

    tree = Tree()
    merge_tree = tree

    if module.has_trait(F.has_part_picked):
        return tree

    if module.has_trait(F.is_pickable):
        merge_tree = Tree()
        tree[module] = merge_tree

    for child in module.get_children(
        direct_only=True, types=(Module, ModuleInterface), include_root=False
    ):
        child_tree = get_pick_tree(child)
        merge_tree.update(child_tree)

    return tree


def update_pick_tree(tree: Tree[Module]) -> tuple[Tree[Module], bool]:
    if not tree:
        return tree, False

    filtered_tree = Tree(
        (k, sub[0])
        for k, v in tree.items()
        if not k.has_trait(F.has_part_picked) and not (sub := update_pick_tree(v))[1]
    )
    if not filtered_tree:
        return filtered_tree, True

    return filtered_tree, False


def check_missing_picks(module: Module):
    # - not skip self pick
    # - no parent with part picked
    # - not specialized
    # - no module children
    # - no parent with picker

    missing = module.get_children_modules(
        types=Module,
        direct_only=False,
        include_root=True,
        # not specialized
        most_special=True,
        # leaf == no children
        f_filter=lambda m: not m.get_children_modules(
            types=Module, f_filter=lambda x: not isinstance(x, F.Footprint)
        )
        and not isinstance(m, F.Footprint)
        # no parent with part picked
        and not try_or(
            lambda: m.get_parent_with_trait(F.has_part_picked),
            default=False,
            catch=KeyErrorNotFound,
        )
        # no parent with picker
        and not try_or(
            lambda: try_or(
                lambda: m.get_parent_with_trait(F.is_pickable),
                default=False,
                catch=KeyErrorNotFound,
            ),
            default=True,
            catch=KeyErrorAmbiguous,
        ),
    )

    if missing:
        no_fp, with_fp = map(
            list, partition(lambda m: m.has_trait(F.has_footprint), missing)
        )

        if with_fp:
            logger.warning(f"No pickers for {indented_container(with_fp)}")
        if no_fp:
            logger.warning(
                f"No pickers and no footprint for {indented_container(no_fp)}."
                "\nATTENTION: These modules will not appear in netlist or pcb."
            )


def find_independent_groups(
    modules: Iterable[Module], solver: Solver
) -> list[set[Module]]:
    """
    Find groups of modules that are independent of each other.
    """
    from faebryk.core.solver.defaultsolver import DefaultSolver

    modules = set(modules)

    if (
        not isinstance(solver, DefaultSolver)
        or (state := solver.reusable_state) is None
    ):
        # Find params aliased to lits
        aliased = EquivalenceClasses[ParameterOperatable]()
        lits = dict[ParameterOperatable, ParameterOperatable.Literal]()
        for e in GraphFunctions(*get_graphs(modules)).nodes_of_type(Is):
            if not e.constrained:
                continue
            aliased.add_eq(*e.operatable_operands)
            if e.get_operand_literals():
                lits[e.get_operand_operatables().pop()] = next(
                    iter(e.get_operand_literals().values())
                )
        for alias_group in aliased.get():
            lits_eq = unique((lits[p] for p in alias_group if p in lits), lambda x: x)
            if not lits_eq:
                continue
            if len(lits_eq) > 1:
                # TODO
                raise ContradictionByLiteral("", [], [], None)
            for p in alias_group:
                lits[p] = lits_eq[0]

        # find params related to each other
        param_eqs = EquivalenceClasses[Parameter]()
        for e in GraphFunctions(*get_graphs(modules)).nodes_of_type(
            ConstrainableExpression
        ):
            if not e.constrained:
                continue
            ps = e.get_operand_parameters(recursive=True)
            ps.difference_update(lits.keys())
            param_eqs.add_eq(*ps)

        # find modules that are dependent on each other
        module_eqs = EquivalenceClasses[Module](modules)
        for p_eq in param_eqs.get():
            p_modules = {
                m
                for p in p_eq
                if (parent := p.get_parent()) is not None
                and isinstance(m := parent[0], Module)
                and m in modules
            }
            module_eqs.add_eq(*p_modules)
        out = module_eqs.get()
        logger.debug(indented_container(out, recursive=True))
        return out

    graphs = EquivalenceClasses()
    graph_to_m = defaultdict[Graph, set[Module]](set)
    for m in modules:
        params = m.get_trait(F.is_pickable_by_type).get_parameters().values()
        new_params = {state.data.mutation_map.map_forward(p).maps_to for p in params}
        m_graphs = get_graphs(new_params)
        graphs.add_eq(*m_graphs)
        for g in m_graphs:
            graph_to_m[g].add(m)

    graph_groups: list[set[Graph]] = graphs.get()
    module_groups = [{m for g in gg for m in graph_to_m[g]} for gg in graph_groups]
    return module_groups


def _list_to_hack_tree(modules: Iterable[Module]) -> Tree[Module]:
    return Tree({m: Tree() for m in modules})


def pick_topologically(
    tree: Tree[Module], solver: Solver, progress: Advancable | None = None
):
    # TODO implement backtracking

    import faebryk.libs.picker.api.picker_lib as picker_lib

    timings = Times(name="pick")
    logger.info(f"Picking {len(tree)} modules")

    explicit_modules = [
        m
        for m in tree.keys()
        if m.has_trait(F.is_pickable_by_part_number)
        or m.has_trait(F.is_pickable_by_supplier_id)
    ]
    logger.info(f"Picking {len(explicit_modules)} explicit parts")
    explicit_parts = picker_lib._find_modules(
        _list_to_hack_tree(explicit_modules), solver
    )
    for m, parts in explicit_parts.items():
        part = parts[0]
        picker_lib.attach_single_no_check(m, part, solver)
        if progress:
            progress.advance()
    if explicit_parts:
        tree, _ = update_pick_tree(tree)

    tree_backup = set(tree.keys())
    _pick_count = len(tree)

    def _get_candidates(_tree: Tree[Module]):
        # with timings.as_global("pre-solve"):
        #    solver.simplify(*get_graphs(tree.keys()))
        try:
            with timings.as_global("new estimates"):
                # Rerun solver for new system
                solver.update_superset_cache(*_tree)
        except Contradiction as e:
            raise PickError(str(e), *_tree.keys())
        with timings.as_global("get candidates"):
            candidates = picker_lib.get_candidates(_tree, solver)
        if LOG_PICK_SOLVE:
            logger.info(
                "Candidates: \n\t"
                f"{'\n\t'.join(f'{m}: {len(p)}' for m, p in candidates.items())}"
            )
        return candidates

    timings.add("setup")

    # Works by looking for each module again for compatible parts
    # If no compatible parts are found,
    #   it means we need to backtrack or there is no solution
    with timings.as_global("parallel slow-pick", context=True):
        if tree:
            logger.info(f"Picking {len(tree)} modules in parallel")
        while tree:
            try:
                candidates = _get_candidates(tree)
            except Contradiction as e:
                # TODO better error, also remove from get_candidates
                raise PickError(
                    "Contradiction in system."
                    "Unfixable due to lack of backtracking." + str(e),
                    *[],
                )
            groups = find_independent_groups(candidates.keys(), solver)
            # pick module with least candidates first
            picked = [
                (
                    m := min(group, key=lambda _m: len(candidates[_m])),
                    candidates[m][0],
                )
                for group in groups
            ]
            logger.info(
                f"Picking {len(groups)} independent groups: "
                f"{indented_container([m for m, _ in picked])}"
            )
            for m, part in picked:
                picker_lib.attach_single_no_check(m, part, solver)
                if progress:
                    progress.advance()

            tree, _ = update_pick_tree(tree)

    if _pick_count:
        logger.info(
            f"Slow-picked parts in {timings.get_formatted('parallel slow-pick')}"
        )

    logger.info(f"Picked complete: picked {_pick_count} parts")
    logger.info("Verify design")
    try:
        solver.update_superset_cache(*tree_backup)
    except Contradiction as e:
        raise PickVerificationError(str(e), *tree_backup) from e


# TODO should be a Picker
def pick_part_recursively(
    module: Module, solver: Solver, progress: Advancable | None = None
):
    pick_tree = get_pick_tree(module)
    if progress:
        progress.set_total(len(pick_tree))
    check_missing_picks(module)

    try:
        pick_topologically(pick_tree, solver, progress)
    # FIXME: This does not get called anymore
    except PickErrorChildren as e:
        failed_parts = e.get_all_children()
        for m, sube in failed_parts.items():
            logger.error(
                f"Could not find pick for {m}:\n {sube.message}\n"
                f"Params:\n{indent(m.pretty_params(solver), prefix=' '*4)}"
            )
        raise



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from cmath import pi
from typing import cast

from faebryk.core.parameter import (
    Add,
    And,
    CanonicalExpression,
    Ceil,
    ConstrainableExpression,
    Cos,
    Difference,
    Divide,
    Expression,
    Floor,
    GreaterOrEqual,
    GreaterThan,
    Implies,
    Intersection,
    IsSubset,
    IsSuperset,
    LessOrEqual,
    LessThan,
    Max,
    Min,
    Multiply,
    Not,
    Or,
    Parameter,
    ParameterOperatable,
    Power,
    Round,
    Sin,
    Sqrt,
    Subtract,
    SymmetricDifference,
    Union,
    Xor,
)
from faebryk.core.solver.algorithm import algorithm
from faebryk.core.solver.mutator import Mutator
from faebryk.core.solver.utils import (
    SolverAllExtended,
    make_lit,
)
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
    QuantityLikeR,
)
from faebryk.libs.sets.sets import as_lit
from faebryk.libs.units import dimensionless, quantity
from faebryk.libs.util import cast_assert

logger = logging.getLogger(__name__)

NumericLiteralR = (*QuantityLikeR, Quantity_Interval_Disjoint, Quantity_Interval)


@algorithm("Constrain within and domain", single=True, terminal=False)
def constrain_within_domain(mutator: Mutator):
    """
    Translate domain and within constraints to parameter constraints.
    """

    for param in mutator.nodes_of_type(Parameter):
        new_param = mutator.mutate_parameter(param, override_within=True, within=None)
        if param.within is not None:
            mutator.utils.subset_to(new_param, param.within, from_ops=[param])
        mutator.utils.subset_to(
            new_param,
            param.domain_set(),
            from_ops=[param],
        )


@algorithm("Alias predicates to true", single=True, terminal=False)
def alias_predicates_to_true(mutator: Mutator):
    """
    Alias predicates to True since we need to assume they are true.
    """

    for predicate in mutator.nodes_of_type(ConstrainableExpression):
        if predicate.constrained:
            new_predicate = cast_assert(
                ConstrainableExpression, mutator.mutate_expression(predicate)
            )
            mutator.utils.alias_to(new_predicate, as_lit(True))
            # reset solver flag
            mutator.predicate_reset_termination(new_predicate)


@algorithm("Canonical literal form", single=True, terminal=False)
def convert_to_canonical_literals(mutator: Mutator):
    """
    - remove units for NumberLike
    - NumberLike -> Quantity_Interval_Disjoint
    - bool -> BoolSet
    - Enum -> P_Set[Enum]
    """

    param_ops = mutator.nodes_of_type(ParameterOperatable, sort_by_depth=True)

    for po in param_ops:
        # Parameter
        if isinstance(po, Parameter):
            mutator.mutate_parameter(
                po,
                units=dimensionless,
                soft_set=make_lit(po.soft_set).to_dimensionless()
                if po.soft_set is not None
                else None,
                within=make_lit(po.within).to_dimensionless()
                if po.within is not None
                else None,
                guess=quantity(po.guess, dimensionless)
                if po.guess is not None
                else None,
                override_within=True,
            )

        # Expression
        elif isinstance(po, Expression):

            def mutate(i: int, operand: SolverAllExtended) -> SolverAllExtended:
                if not ParameterOperatable.is_literal(operand):
                    return operand
                lit = make_lit(operand)
                if isinstance(lit, Quantity_Interval_Disjoint):
                    return lit.to_dimensionless()
                return lit

            # need to ignore existing because non-canonical literals
            # are congruent to canonical
            mutator.mutate_expression_with_op_map(po, mutate, ignore_existing=True)


@algorithm("Canonical expression form", single=True, terminal=False)
def convert_to_canonical_operations(mutator: Mutator):
    """
    Transforms Sub-Add to Add-Add
    ```
    A - B -> A + (-1 * B)
    A / B -> A * B^-1
    A <= B -> B >= A
    A < B -> B > A
    A superset B -> B subset A
    Logic (xor, and, implies) -> Or & Not
    floor/ceil -> round(x -/+ 0.5)
    cos(x) -> sin(x + pi/2)
    sqrt(x) -> x^-0.5
    min(x) -> p, p ss x, p le x
    max(x) -> p, p ss x, p ge x
    ```
    """

    UnsupportedOperations: dict[type[Expression], type[Expression]] = {
        GreaterThan: GreaterOrEqual,
        LessThan: LessOrEqual,
    }

    def c[T: CanonicalExpression](op: type[T], *operands) -> T:
        return mutator.create_expression(
            op, *operands, from_ops=getattr(c, "from_ops", None)
        )

    def curry(e_type: type[CanonicalExpression]) -> type[Expression]:
        def _(*operands):
            operands = [
                make_lit(o) if not isinstance(o, ParameterOperatable) else o
                for o in operands
            ]
            return c(e_type, *operands)

        # hack
        return cast(type[Expression], _)

    # CanonicalNumeric
    Add_ = curry(Add)
    Multiply_ = curry(Multiply)
    Power_ = curry(Power)
    # Round_ = curry(Round)
    # Abs_ = curry(Abs)
    # Sin_ = curry(Sin)
    # Log_ = curry(Log)

    # CanonicalLogic
    Or_ = curry(Or)
    Not_ = curry(Not)

    # CanonicalSetic
    # Intersection_ = curry(Intersection)
    Union_ = curry(Union)
    SymmetricDifference_ = curry(SymmetricDifference)

    # CanonicalPredicate
    # GreaterOrEqual_ = curry(GreaterOrEqual)
    # IsSubset_ = curry(IsSubset)
    # Is_ = curry(Is)
    # GreaterThan_ = curry(GreaterThan)

    MirroredExpressions = [
        (
            Add,
            Subtract,
            lambda operands: [operands[0]] + [Multiply_(o, -1) for o in operands[1:]],
        ),
        (
            Multiply,
            Divide,
            lambda operands: [operands[0]] + [Power_(o, -1) for o in operands[1:]],
        ),
        (
            Not,
            And,
            lambda operands: [Or_(*[Not_(o) for o in operands])],
        ),
        (
            Or,
            Implies,
            lambda operands: [Not_(operands[0]), *operands[1:]],
        ),
        (
            Not,
            Xor,
            lambda operands: [
                Or_(Not_(Or_(*operands)), Not_(Or_(*[Not_(o) for o in operands])))
            ],
        ),
        (
            Round,
            Floor,
            lambda operands: [Add_(*operands, -0.5)],
        ),
        (
            Round,
            Ceil,
            lambda operands: [Add_(*operands, 0.5)],
        ),
        (
            Sin,
            Cos,
            lambda operands: [Add_(*operands, pi / 2)],
        ),
        (
            Power,
            Sqrt,
            lambda operands: [*operands, make_lit(0.5)],
        ),
        (
            GreaterOrEqual,
            LessOrEqual,
            lambda operands: list(reversed(operands)),
        ),
        (
            # GreaterThan,
            # TODO
            GreaterOrEqual,
            LessThan,
            lambda operands: list(reversed(operands)),
        ),
        # TODO remove once support for LT/GT
        (
            GreaterOrEqual,
            GreaterThan,
            lambda operands: operands,
        ),
        (
            IsSubset,
            IsSuperset,
            lambda operands: list(reversed(operands)),
        ),
        (
            # A - B - C = A - (B | C)
            # = A & (A ^ (B | C))
            Intersection,
            Difference,
            lambda operands: [
                operands[0],
                SymmetricDifference_(operands[0], Union_(*operands)),
            ],
        ),
    ]

    lookup = {
        Convertible: (Target, Converter)
        for Target, Convertible, Converter in MirroredExpressions
    }

    exprs = mutator.nodes_of_type(Expression, sort_by_depth=True)
    for e in exprs:
        if type(e) in UnsupportedOperations:
            logger.warning(
                f"{type(e)}({e.compact_repr(mutator.print_context)}) not supported "
                f"by solver, converting to {UnsupportedOperations[type(e)]}"
            )

        from_ops = [e]
        # TODO move up, by implementing Parameter Target
        # Min, Max
        if isinstance(e, (Min, Max)):
            p = Parameter(units=e.units)
            mutator.register_created_parameter(p, from_ops=from_ops)
            union = Union(*[mutator.get_copy(o) for o in e.operands])
            mutator.create_expression(
                IsSubset, p, union, from_ops=from_ops, constrain=True
            )
            if isinstance(e, Min):
                mutator.create_expression(GreaterOrEqual, union, p, from_ops=from_ops)
            else:
                mutator.create_expression(GreaterOrEqual, p, union, from_ops=from_ops)
            mutator._mutate(e, p)
            continue

        if type(e) not in lookup:
            continue

        # Rest
        Target, Converter = lookup[type(e)]

        setattr(c, "from_ops", from_ops)
        mutator.mutate_expression(
            e,
            Converter(e.operands),
            expression_factory=Target,
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import re
from collections import defaultdict
from typing import cast

from natsort import natsorted

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.exporters.pcb.kicad.transformer import PCB, PCB_Transformer
from faebryk.libs.exceptions import UserResourceException
from faebryk.libs.library import L
from faebryk.libs.util import duplicates, groupby, md_list

logger = logging.getLogger(__name__)


def attach_random_designators(graph: Graph):
    """
    Sorts nodes by path and then sequentially attaches designators

    This ensures that everything which has a footprint must have a designator.
    """

    nodes = {n for n, _ in GraphFunctions(graph).nodes_with_trait(F.has_footprint)}

    in_use = {
        n.get_trait(F.has_designator).get_designator()
        for n in nodes
        if n.has_trait(F.has_designator)
    }
    pattern = re.compile(r"([A-Z]+)([0-9]+)")

    groups = groupby(
        [(m.group(1), int(m.group(2))) for d in in_use if (m := pattern.match(d))],
        key=lambda x: x[0],
    )

    assigned = defaultdict(
        list,
        {k: cast(list[int], [num for _, num in v]) for k, v in groups.items()},
    )

    def _get_first_hole(used: list[int]):
        s_used = sorted(used)
        for i in range(len(used)):
            if i + 1 != s_used[i]:
                return i + 1
        return len(used) + 1

    nodes_sorted = natsorted(nodes, key=lambda x: x.get_full_name())

    for n in nodes_sorted:
        if n.has_trait(F.has_designator):
            continue
        if not n.has_trait(F.has_designator_prefix):
            prefix = type(n).__name__
            logger.warning(f"Node {prefix} has no designator prefix")
        else:
            prefix = n.get_trait(F.has_designator_prefix).get_prefix()

        next_num = _get_first_hole(assigned[prefix])
        designator = f"{prefix}{next_num}"
        n.add(F.has_designator(designator))

        assigned[prefix].append(next_num)

    no_designator = {n for n in nodes if not n.has_trait(F.has_designator)}
    assert not no_designator

    dupes = duplicates(nodes, lambda n: n.get_trait(F.has_designator).get_designator())
    assert (
        not dupes
    ), f"Duplicate designators found in layout:\n{md_list(dupes, recursive=True)}"


def load_designators(graph: Graph, attach: bool = False) -> dict[L.Node, str]:
    """
    Load designators from attached footprints and attach them to the nodes.
    """

    def _get_reference(fp: PCB.C_pcb_footprint):
        try:
            return fp.propertys["Reference"].value
        except KeyError:
            return None

    def _get_pcb_designator(fp_trait: PCB_Transformer.has_linked_kicad_footprint):
        fp = fp_trait.get_fp()
        if not fp.name:
            return None
        return _get_reference(fp)

    nodes = GraphFunctions(graph).nodes_with_trait(
        PCB_Transformer.has_linked_kicad_footprint
    )

    known_designators = {
        node: ref
        for node, trait in nodes
        if (ref := _get_pcb_designator(trait)) is not None
        and not isinstance(node, F.Footprint)
    }

    if attach:
        if dups := duplicates(known_designators, lambda x: known_designators[x]):
            dups_fmt = {k: [f"`{m}`" for m in v] for k, v in dups.items()}
            raise UserResourceException(
                f"Duplicate designators found in layout:\n"
                f"{md_list(dups_fmt, recursive=True)}"
            )
        for node, designator in known_designators.items():
            node.add(F.has_designator(designator))

    return known_designators



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.module import Module


class has_descriptive_properties(Module.TraitT):
    def get_properties(self) -> dict[str, str]:
        raise NotImplementedError()

    def add_properties(self, properties: dict[str, str]):
        raise NotImplementedError()

    @staticmethod
    def get_from(obj: Module, key: str) -> str | None:
        if not obj.has_trait(has_descriptive_properties):
            return None
        return obj.get_trait(has_descriptive_properties).get_properties().get(key)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
Faebryk samples demonstrate the usage by building example systems.
This particular sample creates a netlist with an led and a nand ic
    that creates some logic.
The goal of this sample is to show how faebryk can be used to iteratively
    expand the specifics of a design in multiple steps.
Thus this is a netlist sample.
Netlist samples can be run directly.
"""

import logging

import faebryk.library._F as F
from faebryk.core.graph import GraphFunctions
from faebryk.core.module import Module
from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class PowerSource(Module):
    power: F.ElectricPower


class XOR_with_NANDS(F.LogicGates.XOR):
    nands = L.list_field(4, lambda: F.LogicGates.NAND(2))

    def __init__(self):
        super().__init__(2)

    def __preinit__(self):
        A = self.inputs[0]
        B = self.inputs[1]

        G = self.nands
        Q = self.outputs[0]

        # ~(a&b)
        q0 = G[0].get_trait(F.LogicOps.can_logic_nand).nand(A, B)
        # ~(a&~b)
        q1 = G[1].get_trait(F.LogicOps.can_logic_nand).nand(A, q0)
        # ~(~a&b)
        q2 = G[2].get_trait(F.LogicOps.can_logic_nand).nand(B, q0)
        # (a&~b) o| (~a&b)
        q3 = G[3].get_trait(F.LogicOps.can_logic_nand).nand(q1, q2)

        Q.connect(q3)


class App(Module):
    def __preinit__(self) -> None:
        super().__preinit__()

        # levels
        on = F.Logic()
        off = F.Logic()

        # power
        power_source = PowerSource()

        # alias
        power = power_source.power

        # logic
        logic_in = F.Logic()
        logic_out = F.Logic()

        xor = F.LogicGates.XOR(2)
        logic_out.connect(xor.get_trait(F.LogicOps.can_logic_xor).xor(logic_in, on))

        # led
        led = F.LEDIndicator(use_mosfet=True)
        led.power_in.connect(power)

        # application
        switch = F.Switch(F.Logic)()

        logic_in.connect_via(switch, on)

        # bring logic signals into electrical domain
        e_in = logic_in.specialize(F.ElectricLogic())
        e_out = logic_out.specialize(F.ElectricLogic())
        e_on = on.specialize(F.ElectricLogic())
        e_off = off.specialize(F.ElectricLogic())
        e_in.reference.connect(power)
        e_out.reference.connect(power)
        e_on.reference.connect(power)
        e_off.reference.connect(power)
        e_in.set_weak(on=False, owner=self)
        e_on.set(on=True)
        e_off.set(on=False)

        e_out.connect(led.logic_in)

        nxor = xor.specialize(XOR_with_NANDS())
        battery = power_source.specialize(F.Battery())

        el_switch = switch.specialize(F.Switch(F.ElectricLogic)())
        e_switch = F.Switch(F.Electrical)()
        e_switch = el_switch.specialize(
            e_switch,
            matrix=[(e, el.line) for e, el in zip(e_switch.unnamed, el_switch.unnamed)],
        )
        e_switch.add(
            F.has_explicit_part.by_supplier(
                "C318884",
                pinmap={
                    "1": e_switch.unnamed[0],
                    "2": e_switch.unnamed[0],
                    "3": e_switch.unnamed[1],
                    "4": e_switch.unnamed[1],
                },
            )
        )

        # build graph
        for c in [
            led,
            switch,
            battery,
            e_switch,
        ]:
            self.add(c)

        # parametrizing
        for _, t in GraphFunctions(self.get_graph()).nodes_with_trait(
            F.ElectricLogic.has_pulls
        ):
            for pull_resistor in (r for r in t.get_pulls() if r):
                pull_resistor.resistance.constrain_subset(
                    L.Range.from_center_rel(100 * P.kohm, 0.05)
                )
        power_source.power.voltage.constrain_subset(
            L.Range.from_center_rel(3 * P.V, 0.05)
        )
        led.led.led.brightness.constrain_subset(
            TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
        )

        # packages single nands as explicit IC
        nand_ic = F.TI_CD4011BE()
        for ic_nand, xor_nand in zip(nand_ic.gates, nxor.nands):
            xor_nand.specialize(ic_nand)

        # connect power to IC
        nand_ic.power.connect(power_source.power)

        self.add(nand_ic)

        # TODO: remove when we have a LED picker
        led.led.led.add(F.has_explicit_part.by_supplier("C965802"))
        led.led.led.forward_voltage.alias_is(2.4 * P.V)
        led.led.led.max_brightness.alias_is(435 * P.millicandela)
        led.led.led.max_current.alias_is(20 * P.mA)
        led.led.led.color.alias_is(F.LED.Color.YELLOW)

        # TODO remove when we have a battery picker
        battery.add(
            F.has_explicit_part.by_supplier(
                "C5239862",
                pinmap={
                    "1": battery.power.lv,
                    "2": battery.power.hv,
                },
            )
        )
        battery.voltage.alias_is(3 * P.V)

        # TODO remove when we have a MOSFET picker
        mosfet = led.get_children_modules(types=F.MOSFET).pop()
        mosfet.add(F.has_explicit_part.by_supplier("C5364305"))
        mosfet.channel_type.alias_is(F.MOSFET.ChannelType.N_CHANNEL)
        mosfet.saturation_type.alias_is(F.MOSFET.SaturationType.ENHANCEMENT)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import Callable

from faebryk.library.is_optional import is_optional

logger = logging.getLogger(__name__)


class is_optional_defined(is_optional.impl()):
    def __init__(
        self, needed: bool, handle_result: Callable[[bool], None] = lambda _: None
    ):
        self.needed = needed
        self.handle_result = handle_result
        super().__init__()

    def is_needed(self) -> bool:
        return self.needed

    def _handle_result(self, needed: bool) -> None:
        self.handle_result(needed)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class Powered_Relay(Module):
    """
    A relay with MOSFET driver, flyback diode and LED indicator.
    """

    relay: F.Relay
    indicator: F.PoweredLED
    flyback_diode: F.Diode
    relay_driver = L.f_field(F.PowerSwitchMOSFET)(lowside=True, normally_closed=False)

    switch_a_nc: F.Electrical
    switch_a_common: F.Electrical
    switch_a_no: F.Electrical
    switch_b_no: F.Electrical
    switch_b_common: F.Electrical
    switch_b_nc: F.Electrical
    enable: F.ElectricLogic
    power: F.ElectricPower

    def __preinit__(self):
        self.connect_interfaces_by_name(self.relay, allow_partial=True)

        self.relay_driver.power_in.connect(self.power)
        self.relay_driver.logic_in.connect(self.enable)
        self.relay_driver.switched_power_out.connect(self.relay.coil_power)

        self.relay.coil_power.lv.connect_via(
            self.flyback_diode, self.relay.coil_power.hv
        )
        self.indicator.power.connect(self.relay_driver.switched_power_out)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Shenzhen_Kinghelm_Elec_KH_BNC75_3511(Module):
    """
    Inner hole BNC connector board-end elbow 3GHz -55℃~+155℃ 75Ω 9.5mm
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    line: F.ElectricSignal
    shield: F.Electrical

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C2837588"})
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Shenzhen Kinghelm Elec",
            DescriptiveProperties.partno: "KH-BNC75-3511",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wmsc.lcsc.com/wmsc/upload/file/pdf/v2/lcsc/2404191039_Shenzhen-Kinghelm-Elec-KH-BNC75-3511_C2837588.pdf"
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.line.line: ["OUT"],
                self.shield: ["2", "3", "4"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



from faebryk.libs.util import FuncDict, FuncSet


def test_func_dict_contains():
    a = FuncDict([(1, 2), (FuncDict, 4), (FuncSet, 5)])
    assert 1 in a
    assert FuncDict in a
    assert FuncSet in a

    assert a[1] == 2
    assert a[FuncDict] == 4
    assert a[FuncSet] == 5

    assert id not in a

    a[id] = 10
    assert a[id] == 10


def test_func_dict_iter():
    items = [(1, 2), (FuncDict, 4), (FuncSet, 5)]
    a = FuncDict(items)
    assert len(a) == len(items)
    assert all(item[0] in a for item in items)

    for item, aa in zip(items, a):
        assert item[0] == aa

    for (ia, iv), (ak, av) in zip(items, a.items()):
        assert ak == ia
        assert av == iv


def test_func_dict_hash_collision():
    a = FuncDict(hasher=lambda _: 1)
    a[1] = "a"
    a[2] = "b"
    assert len(a) == 2
    assert a[1] == "a"
    assert a[2] == "b"


def test_func_dict_backwards_lookup():
    a = FuncDict()
    a[1] = "a"
    a[2] = "b"
    assert a.backwards_lookup("a") == 1
    assert a.backwards_lookup("b") == 2


def test_func_dict_setdefault():
    a = FuncDict()
    assert 1 not in a
    assert a.setdefault(1, "a") == "a"
    assert a[1] == "a"


def test_func_set_contains():
    a = FuncSet([1, 2, 3, FuncDict, FuncSet])
    assert 1 in a
    assert FuncDict in a
    assert FuncSet in a
    assert 4 not in a


def test_iter_func_set():
    items = [1, 2, 3, FuncDict, FuncSet]
    a = FuncSet(items)
    assert len(a) == len(items)
    assert all(item in a for item in items)

    for item, aa in zip(items, a):
        assert item == aa


def test_func_set_hash_collision():
    a = FuncSet(hasher=lambda _: 1)
    a.add(1)
    a.add(2)
    assert len(a) == 2
    assert 1 in a
    assert 2 in a


def test_func_set_discard():
    a = FuncSet((1, 2), hasher=lambda x: x)
    assert 1 in a
    assert 2 in a
    a.discard(1)
    assert 1 not in a
    assert 2 in a



import textwrap
from typing import Iterable

from antlr4 import InputStream

from atopile.front_end import Bob, TypeRef
from atopile.parse import make_parser
from atopile.parser.AtoParser import AtoParser
from faebryk.core.parameter import Expression, Is, Parameter
from faebryk.libs.library import L
from faebryk.libs.sets.quantity_sets import Quantity_Interval
from faebryk.libs.units import P
from faebryk.libs.util import cast_assert


def _parse(src: str) -> AtoParser:
    input = InputStream(src)
    input.name = "test"
    return make_parser(input)


def _build_file(src: str, ref: TypeRef = TypeRef(["App"])) -> L.Module:
    tree = _parse(textwrap.dedent(src)).file_input()
    bob = Bob()
    return cast_assert(L.Module, bob.build_ast(tree, ref))


def _lonely[T](iterable: Iterable[T]) -> T:
    items = list(iterable)
    assert len(items) == 1
    return items[0]


def test_assert_is():
    module = _build_file(
        """
        module App:
            a: voltage
            b: dimensionless
            c: dimensionless
            d: dimensionless
            e: dimensionless
            f: dimensionless

            assert a is 2mV +/- 10%
            assert b is c
            assert d is e is f
        """
    )

    a, b, c, d, e, f = sorted(
        (p for p in module.get_children(False, types=Parameter)),
        key=lambda p: p.get_name(),
    )

    a_is = _lonely(a.operated_on.get_connected_nodes(types=[Expression]))
    assert isinstance(a_is, Is)
    should_be_a, should_be_2_ish_mv = a_is.operands
    assert should_be_a is a
    assert isinstance(should_be_2_ish_mv, Quantity_Interval)
    assert a_is.constrained
    # FIXME: this float comparison might be brittle
    assert Quantity_Interval.from_center_rel(2 * P.mV, 0.1) == should_be_2_ish_mv

    b_c_is = _lonely(b.operated_on.get_connected_nodes(types=[Expression]))
    assert isinstance(b_c_is, Is)
    assert b_c_is.operands == (b, c)
    assert b_c_is.constrained

    d_e_is = _lonely(d.operated_on.get_connected_nodes(types=[Expression]))
    e_f_is = _lonely(f.operated_on.get_connected_nodes(types=[Expression]))
    assert isinstance(d_e_is, Is)
    assert d_e_is.operands == (d, e)
    assert isinstance(e_f_is, Is)
    assert e_f_is.operands == (e, f)
    assert d_e_is.constrained
    assert e_f_is.constrained



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

from faebryk.core.trait import Trait


class has_esphome_config(Trait):
    @abstractmethod
    def get_config(self) -> dict: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
import faebryk.libs.library.L as L

logger = logging.getLogger(__name__)


class is_decoupled(L.Trait.decless()):
    def __init__(self, capacitor: F.Capacitor):
        super().__init__()
        self._capacitor = capacitor

    @property
    def capacitor(self) -> F.Capacitor:
        return self._capacitor



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Crystal_Oscillator(Module):
    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    crystal: F.Crystal
    capacitors = L.list_field(2, F.Capacitor)
    current_limiting_resistor: F.Resistor

    xtal_if: F.XtalIF

    # ----------------------------------------
    #               parameters
    # ----------------------------------------
    # https://blog.adafruit.com/2012/01/24/choosing-the-right-crystal-and-caps-for-your-design/
    # http://www.st.com/internet/com/TECHNICAL_RESOURCES/TECHNICAL_LITERATURE/APPLICATION_NOTE/CD00221665.pdf
    _STRAY_CAPACITANCE = L.Range(1 * P.pF, 5 * P.pF)

    @L.rt_field
    def capacitance(self):
        return (self.crystal.load_capacitance - self._STRAY_CAPACITANCE) * 2

    def __preinit__(self):
        for cap in self.capacitors:
            cap.capacitance.constrain_subset(self.capacitance)

        self.current_limiting_resistor.allow_removal_if_zero()

        # ----------------------------------------
        #                traits
        # ----------------------------------------

        # ----------------------------------------
        #                aliases
        # ----------------------------------------

        # ----------------------------------------
        #                connections
        # ----------------------------------------
        self.crystal.gnd.connect(self.xtal_if.gnd)
        self.crystal.unnamed[0].connect_via(self.capacitors[0], self.xtal_if.gnd)
        self.crystal.unnamed[1].connect_via(self.capacitors[1], self.xtal_if.gnd)

        self.crystal.unnamed[0].connect_via(
            self.current_limiting_resistor, self.xtal_if.xout
        )
        self.crystal.unnamed[1].connect(self.xtal_if.xin)

    @L.rt_field
    def pcb_layout(self):
        from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
        from faebryk.exporters.pcb.layout.heuristic_decoupling import Params
        from faebryk.exporters.pcb.layout.next_to import LayoutNextTo
        from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy

        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type

        self.capacitors[0].add_trait(
            F.has_pcb_layout_defined(
                layout=LayoutNextTo(
                    target=self.crystal.unnamed[0],
                    params=Params(
                        distance_between_pad_edges=1.25, extra_rotation_of_footprint=90
                    ),
                )
            )
        )
        self.capacitors[1].add_trait(
            F.has_pcb_layout_defined(
                layout=LayoutNextTo(
                    target=self.crystal.unnamed[1],
                    params=Params(
                        distance_between_pad_edges=1.25, extra_rotation_of_footprint=90
                    ),
                )
            )
        )

        return F.has_pcb_layout_defined(
            LayoutTypeHierarchy(
                layouts=[
                    LayoutTypeHierarchy.Level(
                        mod_type=F.Crystal,
                        layout=LayoutAbsolute(
                            Point((0, 0, 0, L.NONE)),
                        ),
                    ),
                    # LayoutTypeHierarchy.Level(
                    #    mod_type=F.Capacitor,
                    #    layout=LayoutExtrude(
                    #        base=Point((-3, 0, 0, L.NONE)),
                    #        vector=(0, 6, 180),
                    #        dynamic_rotation=True,
                    #    ),
                    # ),
                    LayoutTypeHierarchy.Level(
                        mod_type=F.Resistor,
                        layout=LayoutAbsolute(Point((-3, -3, 0, L.NONE))),
                    ),
                ]
            ),
        )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.xtal_if.xin, self.xtal_if.xout)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class CapacitorElectrolytic(F.Capacitor):
    pickable = None
    attach_to_footprint = None

    anode: F.Electrical
    cathode: F.Electrical

    def __preinit__(self):
        self.power.hv.connect(self.anode)
        self.power.lv.connect(self.cathode)

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.anode, self.cathode)

    @L.rt_field
    def has_pin_association_heuristic_lookup_table(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power.hv: ["anode", "a"],
                self.power.lv: ["cathode", "c"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )



# pylint: disable=logging-fstring-interpolation

"""
`ato inspect`
"""

import logging
from typing import Annotated

import typer

from atopile import errors
from atopile.address import AddrStr

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class DisplayEntry:
    """
    This class represents the nets that are below the inspected module,
    the equivalent net that is below the context module and
    the individual connections that are made to the inspect net and the context net.
    """

    def __init__(self, net: list[list[AddrStr]]):
        self.inspect_net: list[AddrStr] = net
        self.inspect_consumer: list[AddrStr] = []
        self.context_net: list[AddrStr] = []
        self.context_consumer: list[AddrStr] = []


odd_row = "on grey11 cornflower_blue"
even_row = "on grey15 cornflower_blue"
odd_greyed_row = "on grey11 grey0"
even_greyed_row = "on grey15 grey0"


def inspect(
    entry: Annotated[str | None, typer.Argument()] = None,
    build: Annotated[list[str], typer.Option("--build", "-b", envvar="ATO_BUILD")] = [],
    target: Annotated[
        list[str], typer.Option("--target", "-t", envvar="ATO_TARGET")
    ] = [],
    option: Annotated[
        list[str], typer.Option("--option", "-o", envvar="ATO_OPTION")
    ] = [],
    inspect: str | None = None,
    context: Annotated[
        str | None,
        typer.Option(
            "--context", "-c", help="The context from which to inspect the module"
        ),
    ] = None,
    dump_csv: Annotated[
        str | None,
        typer.Option("--dump-csv", "-d", help="Output the inspection to a CSV file"),
    ] = None,
):
    """
    Utility to inspect what is connected to a component.
    The context sets the boundary where something is considered connected.
    For example: `--inspect rp2040_micro --context rp2040_micro_ki`
    """
    raise errors.UserNotImplementedError("Inspect is not yet implemented.")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum

import faebryk.library._F as F
from faebryk.libs.library import L


class SMDTwoPin(F.Footprint):
    class Type(Enum):
        _01005 = 0
        _0201 = 1
        _0402 = 2
        _0603 = 3
        _0805 = 4
        _1206 = 5
        _1210 = 6
        _1218 = 7
        _2010 = 8
        _2512 = 9

    def __init__(self, type: Type) -> None:
        super().__init__()
        self._type = type

    pins = L.list_field(2, F.Pad)

    class _has_kicad_footprint(F.has_kicad_footprint_equal_ifs):
        def get_kicad_footprint(self) -> str:
            obj = self.obj
            assert isinstance(obj, SMDTwoPin)
            table = {
                SMDTwoPin.Type._01005: "0402",
                SMDTwoPin.Type._0201: "0603",
                SMDTwoPin.Type._0402: "1005",
                SMDTwoPin.Type._0603: "1608",
                SMDTwoPin.Type._0805: "2012",
                SMDTwoPin.Type._1206: "3216",
                SMDTwoPin.Type._1210: "3225",
                SMDTwoPin.Type._1218: "3246",
                SMDTwoPin.Type._2010: "5025",
                SMDTwoPin.Type._2512: "6332",
            }
            return "Resistor_SMD:R_{imperial}_{metric}Metric".format(
                imperial=obj._type.name[1:], metric=table[obj._type]
            )

    kicad_footprint: _has_kicad_footprint
    equal_pins: F.has_equal_pins_in_ifs
    attach_via_pinmap: F.can_attach_via_pinmap_equal



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import pytest

from faebryk.core.cpp import Graph
from faebryk.core.node import Node
from faebryk.core.parameter import (
    Add,
    And,
    Difference,
    Divide,
    Expression,
    Intersection,
    Is,
    Logic,
    Multiply,
    Or,
    Parameter,
    ParameterOperatable,
    Subtract,
    Union,
    Xor,
)
from faebryk.core.solver.algorithm import algorithm
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.mutator import (
    MutationMap,
    MutationStage,
    Mutator,
    Transformations,
)
from faebryk.core.solver.utils import (
    Associative,
    ContradictionByLiteral,
    FullyAssociative,
    MutatorUtils,
)
from faebryk.libs.library import L
from faebryk.libs.logging import rich_to_string
from faebryk.libs.units import P
from faebryk.libs.util import cast_assert, times

logger = logging.getLogger(__name__)


def _create_letters(
    n: int,
) -> tuple[ParameterOperatable.ReprContext, list[Parameter], Graph]:
    context = ParameterOperatable.ReprContext()

    out = []

    class App(Node):
        def __preinit__(self) -> None:
            for _ in range(n):
                p = Parameter()
                name = p.compact_repr(context)
                self.add(p, name)
                out.append(p)

    app = App()
    return context, out, app.get_graph()


@pytest.mark.parametrize(
    "op",
    [
        Add,
        Multiply,
        Subtract,
        Divide,
        And,
        Or,
        Xor,
        Union,
        Intersection,
        Difference,
    ],
)
def test_flatten_associative(op: type[Expression]):
    def flatten(op):
        return MutatorUtils.flatten_associative(op, lambda _, __: True)

    if issubclass(op, Logic):
        domain = L.Domains.BOOL()
    else:
        domain = L.Domains.Numbers.REAL()

    A, B, C, D, E = times(5, lambda: Parameter(domain=domain))

    to_flatten = op(op(A, B), C, op(D, E))
    res = flatten(to_flatten)

    if not issubclass(op, Associative):
        assert len(res.destroyed_operations) == 0
        assert set(res.extracted_operands) == set(to_flatten.operands)
        return

    if not issubclass(op, FullyAssociative):
        assert set(res.extracted_operands) & {A, B, C}
        assert not set(res.extracted_operands) & {D, E}
        assert len(res.destroyed_operations) == 1
        return

    assert set(res.extracted_operands) == {A, B, C, D, E}
    assert len(res.destroyed_operations) == 2


def test_mutator_no_graph_merge():
    p0 = Parameter(units=P.V)
    p1 = Parameter(units=P.A)
    p2 = Parameter(units=P.W)
    alias = p2.alias_is(p0 * p1)

    p3 = Parameter(units=P.V)

    context = ParameterOperatable.ReprContext()

    @algorithm("")
    def algo(mutator: Mutator):
        pass

    mutator = Mutator(
        MutationMap.identity(p0.get_graph(), print_context=context),
        algo=algo,
        iteration=0,
        terminal=True,
    )
    p0_new = cast_assert(Parameter, mutator.get_copy(p0))
    p3_new = cast_assert(Parameter, mutator.get_copy(p3))
    alias_new = cast_assert(Is, mutator.get_copy(alias))

    G = p0.get_graph()
    G_new = p0_new.get_graph()

    assert G is not G_new
    assert alias_new.get_graph() is G_new
    assert p3_new.get_graph() is not G_new
    assert cast_assert(Parameter, mutator.get_mutated(p1)).get_graph() is G_new


def test_get_expressions_involved_in():
    A = Parameter()
    B = Parameter()

    E1 = A + B

    res = MutatorUtils.get_expressions_involved_in(E1)
    assert res == set()

    E2 = E1 + A

    res = MutatorUtils.get_expressions_involved_in(E1)
    assert res == {E2}

    E3 = E2 + B

    res = MutatorUtils.get_expressions_involved_in(E1)
    assert res == {E2, E3}

    res = MutatorUtils.get_expressions_involved_in(E2)
    assert res == {E3}

    res = MutatorUtils.get_expressions_involved_in(E2, up_only=False)
    assert res == {E1, E3}

    res = MutatorUtils.get_expressions_involved_in(E2, up_only=False, include_root=True)
    assert res == {E1, E2, E3}


def test_get_correlations_basic():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    # Create correlations between parameters
    o = A.alias_is(B)  # A and B are correlated through an Is expression

    # Create an expression with correlated operands
    expr = Add(A, B, C)

    # Test correlations
    correlations = list(MutatorUtils.get_correlations(expr))

    # We expect A and B to be correlated
    assert len(correlations) == 1

    # Unpack the correlation
    op1, op2, overlap_exprs = correlations[0]

    # Check that the correlated operands are A and B
    assert {op1, op2} == {A, B}
    assert overlap_exprs == {o}


def test_get_correlations_nested_uncorrelated():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    o = A.alias_is(B)  # A and B are correlated through an Is expression
    inner = A + B
    expr = inner + C
    correlations = list(MutatorUtils.get_correlations(expr))
    inner_correlations = list(MutatorUtils.get_correlations(inner))

    # no correlations between C and (A + B)
    assert not correlations

    assert len(inner_correlations) == 1
    op1, op2, overlap_exprs = inner_correlations[0]
    assert {op1, op2} == {A, B}
    assert overlap_exprs == {o}


def test_get_correlations_nested_correlated():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    o = A.alias_is(B)  # A and B are correlated through an Is expression
    inner = A + C
    expr = inner + B
    correlations = list(MutatorUtils.get_correlations(expr))
    inner_correlations = list(MutatorUtils.get_correlations(inner))

    # no correlations between C and A
    assert not inner_correlations

    assert len(correlations) == 1
    op1, op2, overlap_exprs = correlations[0]
    assert {op1, op2} == {inner, B}
    assert overlap_exprs == {o}


def test_get_correlations_self_correlated():
    A = Parameter()
    E = A + A
    correlations = list(MutatorUtils.get_correlations(E))
    assert len(correlations) == 1
    op1, op2, overlap_exprs = correlations[0]
    assert {op1, op2} == {A}
    assert not overlap_exprs


def test_get_correlations_shared_predicates():
    A = Parameter()
    B = Parameter()

    E = A + B

    correlations = list(MutatorUtils.get_correlations(E))
    assert not correlations

    E2 = Is(A * B, L.Range(0, 10))

    correlations = list(MutatorUtils.get_correlations(E))
    assert not correlations

    E2.constrain()

    correlations = list(MutatorUtils.get_correlations(E))
    assert len(correlations) == 1

    op1, op2, overlap_exprs = correlations[0]
    assert {op1, op2} == {A, B}
    assert overlap_exprs == {E2}


def test_get_correlations_correlated_regression():
    A = Parameter()
    B = Parameter()

    A.alias_is(L.Range(5, 10))
    B.alias_is(L.Range(10, 15))

    # correlate
    o = B.alias_is(A + 5)

    a_neg = A * -1
    E = B + a_neg

    correlations = list(MutatorUtils.get_correlations(E))
    assert len(correlations) == 1

    op1, op2, overlap_exprs = correlations[0]
    assert {op1, op2} == {a_neg, B}
    assert overlap_exprs == {o}


def test_mutation_map_compressed_mapping_forwards_identity():
    context, variables, graph = _create_letters(3)

    mapping = MutationMap.identity(graph, print_context=context)

    f = mapping.compressed_mapping_forwards
    assert {k: v.maps_to for k, v in f.items()} == {v: v for v in variables}


def test_mutation_map_compressed_mapping_backwards_identity():
    context, variables, graph = _create_letters(3)

    mapping = MutationMap.identity(graph, print_context=context)

    b = mapping.compressed_mapping_backwards
    assert b == {v: [v] for v in variables}


def test_mutation_map_compressed_mapping_backwards_copy():
    context, variables, graph = _create_letters(3)

    mapping = MutationMap.identity(graph, print_context=context)

    _, variables_new, graph_new = _create_letters(3)

    mapping_new = mapping.extend(
        MutationStage(
            algorithm="Test",
            iteration=0,
            print_context=mapping.output_print_context,
            transformations=Transformations(
                input_print_context=mapping.output_print_context,
                mutated=dict(zip(variables, variables_new)),
                copied=set(variables),
            ),
        )
    )

    b = mapping_new.compressed_mapping_backwards
    expected = {v_new: [v] for v, v_new in zip(variables, variables_new)}
    assert b == expected


def test_mutation_map_compressed_mapping_backwards_mutate():
    context, variables, graph = _create_letters(3)

    mapping = MutationMap.identity(graph, print_context=context)

    _, variables_new, graph_new = _create_letters(3)

    mapping_new = mapping.extend(
        MutationStage(
            algorithm="Test",
            iteration=0,
            print_context=mapping.output_print_context,
            transformations=Transformations(
                input_print_context=mapping.output_print_context,
                mutated=dict(zip(variables, variables_new)),
            ),
        )
    )

    b = mapping_new.compressed_mapping_backwards
    expected = {v_new: [v] for v, v_new in zip(variables, variables_new)}
    assert b == expected


def test_mutation_map_non_copy_mutated_identity():
    context, variables, graph = _create_letters(3)

    mapping = MutationMap.identity(graph, print_context=context)

    res = mapping.non_trivial_mutated_expressions
    assert res == set()


def test_mutation_map_non_copy_mutated_mutate():
    context, variables, graph = _create_letters(3)

    mapping = MutationMap.identity(graph, print_context=context)

    _, variables_new, graph_new = _create_letters(3)

    mapping_new = mapping.extend(
        MutationStage(
            algorithm="Test",
            iteration=0,
            print_context=mapping.output_print_context,
            transformations=Transformations(
                input_print_context=mapping.output_print_context,
                mutated=dict(zip(variables, variables_new)),
            ),
        )
    )

    res = mapping_new.non_trivial_mutated_expressions
    assert res == set()


def test_mutation_map_non_copy_mutated_mutate_expression():
    context, variables, graph = _create_letters(2)
    op = Add(*variables)

    mapping = MutationMap.identity(graph, print_context=context)

    _, variables_new, graph_new = _create_letters(2)
    op_new = Multiply(*variables_new)

    mapping_new = mapping.extend(
        MutationStage(
            algorithm="Test",
            iteration=0,
            print_context=mapping.output_print_context,
            transformations=Transformations(
                input_print_context=mapping.output_print_context,
                mutated=dict(zip(variables, variables_new)) | {op: op_new},  # type: ignore
            ),
        )
    )

    res = mapping_new.non_trivial_mutated_expressions
    assert res == {op_new}


def test_mutation_map_submap():
    context, variables, graph = _create_letters(2)
    op = Add(*variables)

    mapping = MutationMap.identity(graph, print_context=context)

    _, variables_new, graph_new = _create_letters(2)
    op_new = Multiply(*variables_new)

    mapping_new = mapping.extend(  # noqa: F841
        MutationStage(
            algorithm="Test",
            iteration=0,
            print_context=mapping.output_print_context,
            transformations=Transformations.identity(
                graph,
                input_print_context=mapping.output_print_context,
            ),
        )
    )
    mapping_new2 = mapping.extend(  # noqa: F841
        MutationStage(
            algorithm="Test",
            iteration=0,
            print_context=mapping.output_print_context,
            transformations=Transformations(
                input_print_context=mapping.output_print_context,
                mutated=dict(zip(variables, variables_new)) | {op: op_new},  # type: ignore
            ),
        )
    )

    # TODO


def test_traceback_filtering_chain():
    context, variables, graph = _create_letters(3)
    A, B, C = variables

    E = A + B
    E2 = E + A

    solver = DefaultSolver()
    out = solver.simplify_symbolically(E2, print_context=context, terminal=False)

    E2_new = out.data.mutation_map.map_forward(E2).maps_to
    assert E2_new
    tb = out.data.mutation_map.get_traceback(E2_new)
    logger.info(tb.filtered())


def test_traceback_filtering_tree():
    context, variables, graph = _create_letters(3)
    A, B, C = variables

    B.constrain_subset(L.Range(0, 10))
    C.constrain_subset(L.Range(5, 15))

    A.constrain_subset(B)
    A.constrain_subset(C)

    solver = DefaultSolver()
    out = solver.simplify_symbolically(A, print_context=context, terminal=True)

    A_new = out.data.mutation_map.map_forward(A).maps_to
    assert A_new
    tb = out.data.mutation_map.get_traceback(A_new)
    logger.info(rich_to_string(tb.filtered().as_rich_tree()))

    # A{S|([5, 10])} <-
    #  CONSTRAINED[Transitive subset]  <- A{S|([0, ∞])}
    #   MUTATED[Constrain within]  <- A
    #    MUTATED[Canonical literal]  <- A:  *46E8.A


def test_contradiction_message_subset():
    context, variables, graph = _create_letters(1)
    (A,) = variables

    A.constrain_subset(L.Range(6, 7))
    A.alias_is(L.Range(4, 5))

    solver = DefaultSolver()

    with pytest.raises(ContradictionByLiteral, match="is lit not subset of ss lits"):
        solver.simplify_symbolically(A, print_context=context, terminal=True)


def test_contradiction_message_superset():
    context, variables, graph = _create_letters(1)
    (A,) = variables

    A.constrain_superset(L.Range(0, 10))
    A.alias_is(L.Range(4, 5))

    solver = DefaultSolver()

    with pytest.raises(
        ContradictionByLiteral, match="Contradiction: Incompatible literal subsets"
    ):
        solver.simplify_symbolically(A, print_context=context, terminal=True)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import TYPE_CHECKING

from faebryk.core.module import Module

if TYPE_CHECKING:
    from faebryk.library.Footprint import Footprint


class has_footprint(Module.TraitT):
    @abstractmethod
    def get_footprint(self) -> "Footprint": ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class INA228_ReferenceDesign(Module):
    """
    TODO: add description
    """

    class ShuntedElectricPower(Module):
        power_in: F.ElectricPower
        power_out: F.ElectricPower
        shunt_sense: F.DifferentialPair

        shunt: F.Resistor

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(self.power_in, self.power_out)

        def __init__(self, lowside: bool = False, filtered: bool = False):
            super().__init__()
            self._lowside = lowside
            self._filtered = filtered

        def __preinit__(self):
            self.shunt_sense.p.line.connect_via(self.shunt, self.shunt_sense.n.line)
            # TODO: minus voltagedrop over shunt
            self.power_in.voltage.alias_is(self.power_out.voltage)
            if self._lowside:
                self.power_in.lv.connect_via(self.shunt, self.power_out.lv)
                self.power_in.hv.connect(self.power_out.hv)
            else:
                self.power_in.hv.connect_via(self.shunt, self.power_out.hv)
                self.power_in.lv.connect(self.power_out.lv)

            if self._filtered:
                raise NotImplementedError
            # TODO: add filter
            #    filter_cap = self.add(F.Capacitor())
            #    filter_resistors = L.list_field(2, F.Resistor)
            #
            #    filter_cap.capacitance.constrain_subset(
            #        L.Range.from_center_rel(0.1 * P.uF, 0.01)
            #    )
            #    filter_cap.max_voltage.constrain_subset(
            #        L.Range.from_center_rel(170 * P.V, 0.01)
            #    )
            #    for res in filter_resistors:
            #        res.resistance.constrain_subset(10 * P.kohm)
            # TODO: auto calculate, see: https://www.ti.com/lit/ug/tidu473/tidu473.pdf

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    ina288: F.INA228

    power_load: F.ElectricPower
    power_source: F.ElectricPower

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def can_bridge(self):
        (F.can_bridge_defined(self.power_load, self.power_source))

    def __init__(self, filtered: bool = False, lowside: bool = False):
        super().__init__()
        self._filtered = filtered
        self._lowside = lowside

    def __preinit__(self):
        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        shunted_power = self.add(
            self.ShuntedElectricPower(lowside=self._lowside, filtered=self._filtered)
        )
        shunted_power.shunt.resistance.constrain_subset(
            L.Range.from_center_rel(15 * P.mohm, 0.01)
        )
        shunted_power.shunt.max_power.constrain_subset(
            L.Range.from_center_rel(2 * P.W, 0.01)
        )
        # TODO: calculate according to datasheet p36

        # ----------------------------------------
        #              connections
        # ----------------------------------------
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)

        self.power_load.connect_via(shunted_power, self.power_source)
        self.ina288.bus_voltage_sense.line.connect(self.power_load.hv)

        self.ina288.shunt_input.connect(shunted_power.shunt_sense)

        # decouple power rail
        self.ina288.power.get_trait(F.can_be_decoupled).decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(0.1 * P.uF, 0.01))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
import faebryk.libs.library.L as L
from faebryk.core.module import Module
from faebryk.libs.units import P


class Battery(Module):
    voltage = L.p_field(
        units=P.V,
        soft_set=L.Range(0 * P.V, 100 * P.V),
        likely_constrained=True,
    )
    capacity = L.p_field(
        units=P.Ah,
        soft_set=L.Range(100 * P.mAh, 100 * P.Ah),
        likely_constrained=True,
    )

    power: F.ElectricPower

    def __preinit__(self) -> None:
        self.power.voltage.constrain_subset(self.voltage)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(self.power)

    designator = L.f_field(F.has_designator_prefix)("BAT")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F


class is_pickable_by_supplier_id(F.is_pickable.decless()):
    class Supplier(Enum):
        LCSC = auto()

    def __init__(self, supplier_part_id: str, supplier: Supplier = Supplier.LCSC):
        super().__init__()
        self._supplier_part_id = supplier_part_id
        self._supplier = supplier

    def get_supplier_part_id(self) -> str:
        return self._supplier_part_id

    def get_supplier(self) -> Supplier:
        return self._supplier
